# Copyright 2019-2021 Cambridge Quantum Computing
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pytket.circuit import Circuit, Bit  # type: ignore
from pytket.utils import (
    QubitPauliOperator,
    expectation_from_counts,
)
from pytket.pauli import QubitPauliString  # type: ignore
from pytket.backends.backendresult import BackendResult
import copy
from sympy import Symbol  # type: ignore
from typing import Iterable, Dict, Union, Tuple, List
from collections import OrderedDict
from numpy import ndarray


class SymbolsDict(object):
    """
    A helper class for standardising interfacing with Circuit Symbolics in qermit.
    Methods take different containers that hold some kind of symbols representation
    and return a SymbolsDict object.
    Methods access self._symbolic_map or use other accessors to modify or add new
    symbols.
    """

    def __init__(self):
        """
        Default constructor, creates an empty OrderedDict() object for future symbols to be added to.
        """
        self._symbolic_map: Dict[Symbol, Union[None, float]] = OrderedDict()

    @classmethod
    def symbols_from_circuit(cls, circuit: Circuit) -> "SymbolsDict":
        """
        Given a pytket Circuit, returns a SymbolsDict object capturing
        given circuits free symbols.

        :param circuit: Pytket circuit with potential symbols.
        :type circuit: Circuit

        """
        mit_symbols = cls()
        for sym in circuit.free_symbols():
            mit_symbols.add_symbol(sym)
        return mit_symbols

    @classmethod
    def symbols_from_dict(
        cls, symbol_dict: Dict[Symbol, Union[None, float]]
    ) -> "SymbolsDict":
        """
        Assigns to mit_symbols attribute _symbolic_map straight from passed dictionary
        of Symbol to None/float.

        :param symbol_dict: Dictionary from Circuit symbolics to values.
        :type symbol_dict: Dict[Symbol, Union[None, float]

        """
        mit_symbols = cls()
        mit_symbols._symbolic_map = symbol_dict
        return mit_symbols

    @classmethod
    def symbols_from_list(
        cls, symbols_list: Iterable[Union[Symbol, str]]
    ) -> "SymbolsDict":
        """
        Adds all symbols (or string representing Symbol) as dict entries with no value.

        :param symbols_list: A list of strings representing Symbols or Symbols.
        :type symbols_list: Iterable[Union[Symbol, str]

        """
        mit_symbols = cls()
        for sym in symbols_list:
            mit_symbols.add_symbol(sym)
        return mit_symbols

    @property
    def symbols_list(self) -> Iterable[Symbol]:
        """
        Returns all symbols held in dictionary of symbols in SymbolsDict object.

        :return: Iterable containing all keys from _symbolic_map, i.e. all Symbols
        :rtype: Iterable[Symbol]
        """
        for s in self._symbolic_map.keys():
            yield s

    def add_symbol(self, symbol: Union[str, Symbol]):
        """
        Adds any passed Symbol (in string form or sympy Symbol type) as a key to dictionary with None value assigned.

        :param symbol: Symbol to be added to self._symbolic_map
        :type symbol: Union[str, Symbol]
        """
        if isinstance(symbol, str):
            sym = Symbol(symbol)
            self._symbolic_map[sym] = None
        elif isinstance(symbol, Symbol):
            self._symbolic_map[symbol] = None
        else:
            msg = f"""
            Argument symbol is of invalid type: {type(symbol)}.
            """
            raise TypeError(msg)

    def get_symbolic_map(self, symbol_values: ndarray) -> Dict[Symbol, float]:
        """
        Assigns given values in parameters to keys in self._symbolic_map in order, for a new
        dictionary object. Returns just this dictionary type.

        :param symbol_values: Ordered values to match to ordered keys for new dict object.
        :type symbol_values: ndarray

        :return: New dict object mapping symbol to value.
        :rtype: Dict[Symbol, float]
        """
        _map = {}
        for symbol, value in zip(self._symbolic_map.keys(), symbol_values):
            _map[symbol] = value
        return _map

    def set_values(self, symbol_values: ndarray):
        """
        Assigns given values in parameters to keys in self._symbolic_map in order, for a new
        dictionary object.

        :param symbol_values: Orderd values to match to ordered keys for new dict object.
        :type symbol_values: ndarray

        :return: New dict object mapping symbol to value.
        :rtype: Dict[Symbol, float]
        """
        _map = {}
        for symbol, value in zip(self._symbolic_map.keys(), symbol_values):
            _map[symbol] = value
        self._symbolic_map = _map

    def add_value(self, symbol: Symbol, value: float):
        """
        Assigns value to self._symbolic_map[symbol]. If symbol not in object then throws an error.

        :param symbol: Symbol to have value assigned.
        :type symbol: Symbol
        :param value: Value to assign to symbol.
        :type value: float
        """
        if symbol in self._symbolic_map:
            self._symbolic_map[symbol] = value
        else:
            raise ValueError("Symbol {} not in object.".format(symbol))

    def __str__(self):
        return f"<SymbolsDict::{len(self._symbolic_map)}>"

    def __repr__(self):
        return str(self)


class MeasurementCircuit(object):
    """
    Stores a single measurement circuit that captures one or multiple observable estimations
    for some Ansatz Circuit.
    """

    def __init__(self, symbolic_circuit: Circuit, symbols: SymbolsDict = None):
        """
        Stores information required to instantiate any MeasurementCircuit with parameterised symbols.

        :param symbolic_circuit: Measurement circuit, may or may not have symbolics.
        :type symbolic_circuit: Circuit
        :param symbols: SymbolsDict object holding symbols and values for all symbols in Circuit. Default none if circuit not symbolic.
        :type symbols: SymbolsDict
        """
        self._symbolic_circuit: Circuit = symbolic_circuit
        if symbols is None:
            self._symbols: SymbolsDict = SymbolsDict.symbols_from_circuit(
                symbolic_circuit
            )
        elif isinstance(symbols, SymbolsDict):
            self._symbols = symbols
        else:
            raise ValueError("Passed symbols object of incorrect type.")

    @property
    def circuit(self) -> Circuit:
        """
        Returns measurement circuit stored in oracle.

        :return: Circuit in oracle
        :rtype: Circuit
        """
        return self._symbolic_circuit

    @property
    def symbols(self) -> Tuple[Symbol, ...]:
        """
        Converts symbols_list property held in SymbolsDict to a tuple and returns it.

        :return: All Symbols in object
        :rtype: Tuple[List[Symbol]]
        """
        return tuple(self._symbols.symbols_list)

    def get_parametric_circuit(self) -> Circuit:
        """
        Substitutes parameters held in SymbolDict into copy of circuit and returns.

        :return: Substituted circuit
        :rtype: Circuit
        """
        _circuit = self._symbolic_circuit.copy()
        _circuit.symbol_substitution(self._symbols._symbolic_map)
        return _circuit


MeasurementInfo = Tuple[QubitPauliString, List[Bit], bool]


class ObservableTracker:
    """
    Stores all measurement circuits required to get observable expectations for each
    QubitPauliString in a given QubitPauliOperator.
    """

    def __init__(self, qubit_pauli_operator: QubitPauliOperator = QubitPauliOperator()):
        """
        Default constructor, creates an empty dict object for mapping QubitPauliStrings to measurement circuits
        and the qubits measured to get expectation, along with an empty list for storing measurement circuits and
        a list for storing partitions.

        :param qubit_pauli_operator: QubitPauliOperator for which given ObservableTracker is expected
            to retain measurement circuits all QubitPauliString keys for before any Backend execution.
        :type qubit_pauli_strings: List[QubitPauliString]
        """
        self._qubit_pauli_operator = qubit_pauli_operator
        # indices being inde in measurement circuits
        self._qps_to_indices: Dict[
            QubitPauliString, List[Tuple[int, List[Bit], bool]]
        ] = dict()
        for k in self._qubit_pauli_operator._dict.keys():
            self._qps_to_indices[k] = list()
        self._measurement_circuits: List[MeasurementCircuit] = list()
        self._partitions: List[QubitPauliString] = list()

    def from_ObservableTracker(to_copy: "ObservableTracker") -> "ObservableTracker":
        """
        Copies each class attribute from to_copy to self. Returns self.

        :param to_copy: An alternative ObservableTracker for making a copy of.
        :type to_copy: 'ObservableTracker'

        :return: New ObservableTracker object
        """
        new_obj = ObservableTracker(copy.copy(to_copy._qubit_pauli_operator))

        new_obj._qps_to_indices = copy.copy(to_copy._qps_to_indices)
        new_obj._measurement_circuits = copy.copy(to_copy._measurement_circuits)
        new_obj._partitions = copy.copy(to_copy._partitions)
        return new_obj

    def __str__(self):
        return (
            f"<ObservableTracker::{len(self._measurement_circuits)}MeasurementCircuits>"
        )

    def __repr__(self):
        return str(self)

    def modify_coefficients(
        self, new_coefficients: List[Tuple[QubitPauliString, float]]
    ):
        """
        Updates coefficients in held QubitPauliOperator with new coefficients. Each QubitPauliString
        must already be in self._qubit_pauli_operator

        :param new_coefficients: Each Tuple contains a QubitPauliString a new coefficient.
        :type new_coefficients: List[Tuple[QubitPauliString, float]]
        """
        for coeff in new_coefficients:
            if coeff[0] not in self._qubit_pauli_operator._dict:
                raise ValueError(
                    "Given string {} not held in ObservableTracker object.".format(
                        coeff[0]
                    )
                )
            self._qubit_pauli_operator._dict[coeff[0]] = coeff[1]

    def extend_operator(self, new_operator: QubitPauliOperator):
        """
        Extends self._qubit_pauli_operator to include tuples in passed operator.

        :param new_operator: Each QubitPauliString and coefficient added to held operator.
        :type new_operator: QubitPauliOperator
        """
        self._qubit_pauli_operator += new_operator

    def remove_strings(self, strings: List[QubitPauliString]):
        """
        Removes passed qubit pauli strings from held QubitPauliOperator and dict from string to index.

        :param strings: Qubit Pauli Strings no longer required to be measured by ObservableTracker
        :type strings: List[QubitPauliString]
        """
        for qps in strings:
            self._qps_to_indices.pop(qps, None)
            self._qubit_pauli_operator._dict.pop(qps, None)

    @property
    def qubit_pauli_operator(self):
        """
        Returns stored qubit pauli operator

        :return: QubitPauliOperator object stored in class
        :rtype: QubitPauliOperator
        """
        return self._qubit_pauli_operator

    def add_measurement_circuit(
        self, circuit: MeasurementCircuit, measurement_info: List[MeasurementInfo]
    ):
        """
        Adds given measurement circuit to stored _measurement_circuits attribute and for each qubit pauli string and qubits in associated
        strings, updates dictionary between string and its measurement circuit + bit to measure and whether result should be inverted.

        :param circuit: Measurement circuit to run to get results.
        :type circuit: MeasurementCircuit
        :param measurement_info: Each entry contains a QubitPauliString, the bits required to take expectation over in resulting result
            and a bool signifying whether expectation should be inverted when taking result.
        :type measurement_info: List[MeasurementInfo] i.e. List[Tuple[QubitPauliString, List[Bit], bool]]
        """
        self._measurement_circuits.append(circuit)
        index = len(self._measurement_circuits) - 1

        # assume that if strings have the same circuit, they must commute
        self._partitions.append([m[0] for m in measurement_info])
        for s in measurement_info:
            string = s[0]
            bits = s[1]
            invert = s[2]
            if s[0] not in self._qps_to_indices:
                raise ValueError(
                    "ObservableTracker object does not track {}.".format(s[0])
                )
            self._qps_to_indices[string].append((index, bits, invert))

    def get_measurement_circuits(
        self, string: QubitPauliString
    ) -> List[MeasurementCircuit]:
        """
        Returns the measurements required to be run for a single QubitPauliString's expectation.

        :param string: QubitPauliString of interest.
        :type string: QubitPauliString

        :return: Measurement Circuit run to find expection of QubitPauliString for some undefined ansatz circuit.
        :rtype: MeasurementCircuit
        """
        indices = [t[0] for t in self._qps_to_indices[string]]
        circuits = [self._measurement_circuits[i] for i in indices]
        return circuits

    def check_string(self, string: QubitPauliString) -> bool:
        """
        Returns true if given QubitPauliString has a measurement circuit stored in self._measurement_circuits.

        :param string: Operator measurement circuit existence being checked for.
        :type string: QubitPauliString

        :return: True if string has measurement circuit, false if not.
        :rtype: bool
        """
        if string not in self._qps_to_indices:
            return False
        if len(self._qps_to_indices[string]) > 0:
            return True
        else:
            return False

    def get_empty_strings(self) -> List[QubitPauliString]:
        """
        Returns all strings in operator that don't have some assigned MeasurementCircuit.

        :return: Strings that require some MeasurementCircuit to be set
        :rtype: List[QubitPauliString]
        """
        output = []
        for string in self._qubit_pauli_operator._dict:
            if not self.check_string(string):
                output.append(string)
        return output

    @property
    def measurement_circuits(self) -> List[MeasurementCircuit]:
        """
        Returns all measurement circuits aded to ObservableTracker via get_measurement_circuit.

        :return: All measurement circuits held in ObservableTracker self._measurement_circuits attirbute.
        :rtype: List[MeasurementCircuit]
        """
        return self._measurement_circuits

    # TODO: QSE derives expectations differently for measurement circuits
    # Method will need to be renamed once this is suppported
    def get_expectations(self, results: List[BackendResult]) -> QubitPauliOperator:
        """
        For given list of results, returns a QubitPauliOperator giving an expectation for each QubitPauliString
        held in self._qps_to_indices. Expectation derived by taking parity of counts.

        :param results: Result objects to derive counts and then an expectation from.
        :type result: BackendResult

        :return: Expectation for each QubitPauliString in self._qps_to_indices
        :type: QubitPauliOperator
        """
        max_index = len(results) - 1
        results_dict = dict()
        # find expectation for each qubit pauli string stored in dict
        for qps in self._qps_to_indices:
            expectation = 0
            # measure info stores result index of interest, bits of choice and
            # whether result should be inverted
            for measure_info in self._qps_to_indices[qps]:
                result_index = measure_info[0]
                # suggests something has gone wrong with MitEx piping of tasks
                if result_index > max_index:
                    raise ValueError(
                        "Desired index {} greater than max index {} of results.".format(
                            result_index, max_index
                        )
                    )

                result = results[result_index]
                bits = measure_info[1]
                invert = measure_info[2]
                counts = result.get_counts(bits)
                expectation += ((-1) ** invert) * expectation_from_counts(counts)
            # once expectation has been derived from all suitable results, add to dictionary
            coeff = self._qubit_pauli_operator[qps]
            results_dict[qps] = expectation * coeff
        # package in QubitPauliOperator as MitEx uses
        return QubitPauliOperator(results_dict)



# Copyright 2019-2021 Cambridge Quantum Computing
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pytket.backends import Backend
from pytket.backends.backendresult import BackendResult
from typing import List, Tuple, Dict, Set
from pytket import Qubit, Circuit, Bit, OpType
from qermit import (
    MitTask,
    CircuitShots,
)
from qermit.spam.full_transition_tomography import (
    get_full_transition_tomography_circuits,
    calculate_correlation_matrices,
    correct_transition_noise,
    StateInfo,
    CorrectionMethod,
)


def gen_full_tomography_spam_circuits_task(
    backend: Backend, shots: int, qubit_subsets: List[List[Qubit]]
) -> MitTask:
    """Generate MitTask for calibration circuits according to the specified correlation and given backend.

    :param backend: Backend on which the experiments are run.
    :type backend: Backend
    :param qubit_subsets: A list of lists of correlated Nodes of a `Device`.
        Qubits within the same list are assumed to only have SPAM errors correlated
        with each other. Thus to allow SPAM errors between all qubits you should
        provide a single list.  The qubits in `qubit_subsets` must be nodes in the
        backend's associated `Device`.
    :type qubit_subsets: List[List[Qubit]]
    :param shots: An int corresponding to the number of shots of each calibration circuit required.
    :return: A MitTask object, requiring 1 List[CircuitShots] wire and returning (List[CircuitShots], List[StateInfo])
        corresponding to Calibration Circuits and corresponding states.
    :rtype: MitTask
    """

    def task(
        obj, wire: List[CircuitShots]
    ) -> Tuple[List[CircuitShots], List[CircuitShots], List[StateInfo]]:
        if backend.backend_info is None:
            raise ValueError("Backend has no backend_info attribute.")

        if "FullCorrelatedSpamCorrection" in backend.backend_info.misc:
            # check correlations distance
            if (
                backend.backend_info.misc[
                    "FullCorrelatedSpamCorrection"
                ].CorrelatedNodes
                is qubit_subsets
            ):
                return (wire, [], [])

        process_circuit = Circuit()
        for q in [qb for subset in qubit_subsets for qb in subset]:
            process_circuit.add_qubit(q)
        tomo_circuit_states = get_full_transition_tomography_circuits(
            process_circuit, backend, qubit_subsets
        )

        tomo_circuit_shots = [
            CircuitShots(Circuit=c, Shots=shots) for c in tomo_circuit_states[0]
        ]
        return (wire, tomo_circuit_shots, tomo_circuit_states[1])

    return MitTask(
        _label="SPAMFullTomographyCircuits", _n_in_wires=1, _n_out_wires=3, _method=task
    )


def gen_full_tomography_spam_characterisation_task(
    backend: Backend, qubit_subsets: List[List[Qubit]]
) -> MitTask:
    """
    Uses results from device for characterisation circuits to characterise transition matrices
    for different qubit subsets and stores them in backend.

    :param backend: Backend for storing characterisations in.
    :type backend: Backend
    :param qubit_subsets: Subsets of qubits in backend corresponding to different correlated subsets.
    :type qubit_subsets: List[List[Qubit]]
    """

    def task(
        obj, results: List[BackendResult], state_infos: List[StateInfo]
    ) -> Tuple[bool]:
        """
        :param results: Results from characterisation circuits run on backend.
        :type results: List[BackendResult]
        :param state_infos: Corresponding state prepared in the circuit run for each result and qubit to bit map.
        :type state_infos: List[StateInfo]

        :return: bool confirming characterisation complete.
        :rtype: bool
        """
        if len(results) != len(state_infos):
            raise ValueError(
                "SPAM Characterisation requires the same number of prepared states and results."
            )
        if len(results) > 0:
            characterisation = calculate_correlation_matrices(
                results, state_infos, qubit_subsets
            )

            if backend.backend_info is None:
                raise ValueError("Backend has no backend_info attribute.")
            backend.backend_info.misc["FullCorrelatedSpamCorrection"] = characterisation
        return (True,)

    return MitTask(
        _label="SPAMFullCharacterisationCircuits",
        _n_in_wires=2,
        _n_out_wires=1,
        _method=task,
    )


def gen_full_tomography_spam_correction_task(
    backend: Backend, corr_method: CorrectionMethod
) -> MitTask:
    """
    Uses characterisation result held in backend to correct for SPAM noise in passed
    BackendResult objects. Method used to invert SPAM characteriastion matrices
    and correct results given by CorrectionMethod enum.

    :param backend: Device Backend holding SPAM characterisation.
    :type backend: Backend
    :param corr_method: Method used to invert matrices and correct results.
    :type corr_method: CorrectionMethod
    :param characterised: bool passed from characterisation task confirming characterisation is complete.
    :type characterised: bool
    """

    def task(
        obj,
        results: List[BackendResult],
        bit_qb_maps: List[Tuple[Dict[Qubit, Bit], Dict[Bit, Qubit]]],
        characterised: bool,
    ) -> Tuple[List[BackendResult]]:
        """
        :param results: Results from experiment circuits run on backend.
        :type results: List[BackendResult]
        :param bit_qb_maps: Map between Bits measurement outcomes are assigned to and Qubits in each experiment Circuit. Separate dicts for
        results end of and mid-circuit measurements.
        :type bit_qb_maps: List[Tuple[Dict[Qubit, Bit], Dict[Bit, Qubit]]]

        :return: Corrected Results
        :rtype: Tuple[List[BackendResult]]
        """
        if backend.backend_info is None:
            raise ValueError("Backend has no backend_info attribute.")

        if "FullCorrelatedSpamCorrection" in backend.backend_info.misc:
            char = backend.backend_info.misc["FullCorrelatedSpamCorrection"]
        else:
            raise ValueError(
                "'FullCorrelatedSpamCorrection' not characterised for Backend."
            )
        if len(results) != len(bit_qb_maps):
            raise ValueError(
                "Number of experiment results and Qubit to Bit maps do not match."
            )
        corrected_results = [
            correct_transition_noise(r, qbm, char, corr_method)
            for r, qbm in zip(results, bit_qb_maps)
        ]
        return (corrected_results,)

    return MitTask(
        _label="SPAMFullCorrection", _n_in_wires=3, _n_out_wires=1, _method=task
    )


def get_mid_circuit_measure_map(
    circuit: Circuit, used_bits: Set[Bit] = set()
) -> Dict[Bit, Qubit]:
    """
    For each circuit, gets all Measure commands and uses them to construct a dictionary
    between Bit and Qubit.

    :param circuit: Circuit to get dict between Bit measured and Qubit measured on.
    :type circuit: Circuit

    :return: A dict between Bit measured and Qubit measured on.
    :rtype: Dict[Bit, Qubit]
    """
    bit_to_qubit_map = dict()
    for mc in circuit.commands_of_type(OpType.Measure):
        bit = mc.bits[0]
        if bit not in used_bits:
            bit_to_qubit_map[bit] = mc.qubits[0]
    return bit_to_qubit_map


def gen_get_bit_maps_task() -> MitTask:
    """
    Returns a task that takes a list of circuits and returns the circuits, and a map betwen
    each circuit bit and the qubit it is measured on.
    """

    def task(
        obj, circuit_shots: List[CircuitShots]
    ) -> Tuple[List[CircuitShots], List[Tuple[Dict[Qubit, Bit], Dict[Bit, Qubit]]]]:
        """
        :param circuits: Circuits to retrieve bit maps from.
        :type circuits: List[CircuitShots]

        :return: A tuple comprising the original circuits, and each circuits bit map.
        :rtype: Tuple[List[CircuitShots], List[Tuple[Dict[Qubit, Bit], Dict[Bit, Qubit]]]
        """
        bq_maps = []
        for c in circuit_shots:
            qb_map = c[0].qubit_to_bit_map
            # if condition met, implies that mid circuit measurement has ocurred and not accounted for
            # in this case, iterate through circuit commands to get Qubits for all Bits
            if len(qb_map) != len(c[0].bits):
                bq_maps.append(
                    (qb_map, get_mid_circuit_measure_map(c[0], set(qb_map.values())))
                )
            else:
                # else, just invert map for later correction
                bq_maps.append((qb_map, dict()))
        return (circuit_shots, bq_maps)

    return MitTask(
        _label="GetBitQubitMaps", _n_in_wires=1, _n_out_wires=2, _method=task
    )



# Copyright 2019-2021 Cambridge Quantum Computing
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pytket.backends import Backend
from qermit import (
    MitEx,
    MitRes,
    ObservableTracker,
    AnsatzCircuit,
    MitTask,
    ObservableExperiment,
    TaskGraph,
)
import copy
from pytket.pauli import QubitPauliString  # type: ignore
from enum import Enum
import numpy as np
from scipy.optimize import curve_fit  # type: ignore
from typing import List, Tuple, cast
from pytket import Circuit
from pytket.predicates import CompilationUnit  # type: ignore
from pytket.utils import QubitPauliOperator
import matplotlib.pyplot as plt  # type: ignore
from numpy.polynomial.polynomial import Polynomial


class Folding(Enum):
    """Folding techniques used to increase the noise levels.

    :return: Circuit with gate count increased to scale noise. The unitary implemented by the
        circuit has not changed.
    :rtype: Circuit
    """

    # TODO: Understand why self is needed etc
    # TODO: circ does not appear as input in docs
    # TODO Generalise with 'partial folding' to allow for non integer noise scaling
    def circuit(circ: Circuit, noise_scaling: int, *args) -> Circuit:
        """Noise scaling by circuit folding. In this case the folded circuit is of
        the form :math:`CC^{-1}CC^{-1}...C` where :math:`C` is the original circuit. As such noise may be scaled by
        odd integers. The Unitary implemented is unchanged by this process.

        :param circ: Original circuit to be folded.
        :type circ: Circuit
        :param noise_scaling: Factor by which to scale the noise. This must be an odd integer.
        :type noise_scaling: int
        :raises ValueError: Raised if the amount by which the noise should be scaled is not an odd integer.
        :return: Folded circuit implementing identical unitary to the initial circuit.
        :rtype: Circuit
        """

        # Raise if the amount by which the noise should be scaled is not an odd integer
        if (not noise_scaling % 2) or noise_scaling % 1:
            raise ValueError(
                "When performing circuit folding, the noise scaling must be an odd integer."
            )

        folded_circ = circ.copy()
        for _ in range(noise_scaling // 2):

            # Add barrier between circuit and its inverse
            folded_circ.add_barrier(folded_circ.qubits + folded_circ.bits)
            folded_circ.add_circuit(circ.dagger(), folded_circ.qubits, folded_circ.bits)

            folded_circ.add_barrier(folded_circ.qubits + folded_circ.bits)
            folded_circ.add_circuit(circ, folded_circ.qubits, folded_circ.bits)

        return folded_circ

    def gate(circ: Circuit, noise_scaling: float, _allow_approx_fold: bool) -> Circuit:
        """Noise scaling by gate folding. In this case gates :math:`G` are replaced at random
        with :math:`GG^{-1}G` until the number of gates is sufficiently scaled.

        :param circ: Original circuit to be folded.
        :type circ: Circuit
        :param noise_scaling: Factor by which to increase the noise.
        :type noise_scaling: float
        :param _allow_approx_fold: Allows for the noise to be increased by an amount close to that requested, as
            opposed to by exactly the amount requested.
            This is necessary as there are cases where the exact noise scaling cannot be achieved.
            This occurs due to the discrete
            amounts by which the noise can be increased (i.e. the discrete amount by which one gate increases the noise).
        :type _allow_approx_fold: bool
        :raises ValueError: Raised if the requested noise scaling cannot be exactly achieved. This can be
            avoided by appropriately setting _allow_approx_fold.
        :return: Folded circuit implementing identical unitary to the initial circuit.
        :rtype: Circuit
        """

        c_dict = circ.to_dict()
        num_commands = len(c_dict["commands"])

        # Calculate the number of gates that need to be folded
        num_additional_commands = (noise_scaling - 1) * num_commands
        num_folded_commands = int(num_additional_commands // 2)

        true_noise_scaling = 1 + ((num_folded_commands * 2) / num_commands)
        # This check isolates the case where the noise folding that can be achieved is different
        # from that requested. While noise_scaling can be any real value, as the gates increase
        # the noise by discrete values, not all folding values are possible.
        if (not _allow_approx_fold) and (
            abs(noise_scaling - true_noise_scaling) > 0.001
        ):
            raise ValueError(
                "The noise cannot be scaled by the amount inputted. The noise must be scaled by a factor of the form (#gates + 2i)/#gates, where #gates is the number of gates in the compiled circuit, and i is an integer."
            )

        # Choose a random selection of commands to fold. Commands are referenced by their index.
        # These are chosen with replacement.
        commands_to_fold = np.random.choice(
            [i for i in range(num_commands)], num_folded_commands
        )
        # Calculate how many times each individual command needs to be folded
        num_folds = {i: list(commands_to_fold).count(i) for i in range(num_commands)}

        folded_command_list = []
        # For each command, fold the appropriate number of times.
        for command_index in num_folds:

            command = c_dict["commands"][command_index]

            command_circ_dict = c_dict.copy()
            command_circ_dict["commands"] = [command]
            command_circ = Circuit().from_dict(command_circ_dict)

            # Find the inverse of the command
            inverse_command_circ = command_circ.dagger()
            inverse_command_circ_dict = inverse_command_circ.to_dict()
            inverse_command = inverse_command_circ_dict["commands"]

            # Append command and inverse the appropriate number of times.
            folded_command_list.append(command)
            for _ in range(num_folds[command_index]):
                folded_command_list.append(*inverse_command)
                folded_command_list.append(command)

        folded_c_dict = c_dict.copy()
        folded_c_dict["commands"] = folded_command_list
        folded_c = Circuit().from_dict(folded_c_dict)

        return folded_c


def poly_exp_func(x: float, *params) -> float:
    """Definition of poly-exponential function for the purposes of fitting to data

    :param x: Value at which to evaluate function
    :type x: float
    :return: Evaluation of poly-exponential function
    :rtype: float
    """
    # Note that we use a list ending in 0 so that the constant in the polynomial is 0.
    # The constant can then be absorbed into the coefficient of the exponential.
    return params[0] + params[1] * np.exp(np.polyval([*params[2:], 0], x))


def cube_root_func(x: float, *params) -> float:
    """Definition of cube root function for the purposes of fitting to data.

    :param x: Value at which to evaluate cube root function
    :type x: float
    :return: Evaluation of cube root function
    :rtype: float
    """
    y = x + params[2]
    return params[0] + params[1] * np.sign(y) * (np.abs(y)) ** (1 / 3)


class Fit(Enum):
    """Functions to fit to expectation values as they change with noise.

    :return: Extrapolation of expectation values to the zero noise limit.
    :rtype: float
    """

    # TODO Consider adding adaptive exponential extrapolation
    def cube_root(
        self, x: List[float], y: List[float], _show_fit: bool, *args
    ) -> float:
        """Fit data to a cube root function. This is to say a function of the form :math:`a + b(x+c)^{1/3}`.

        :param x: Noise scaling values.
        :type x: List[float]
        :param y: Expectation values.
        :type y: List[float]
        :param _show_fit: Plot data and resulting fitted function.
        :type _show_fit: bool
        :return: Extrapolation of data to zero noise limit using the best fitting cube root function.
        :rtype: float
        """

        # Fit data to cube root function
        vals = curve_fit(cube_root_func, x, y, p0=[0, 1, 0], maxfev=100000)

        # Evaluate fitted function at 0
        fit_to_zero = cube_root_func(0, *vals[0])

        # Plot fitted function and data
        if _show_fit:

            fit_x = np.linspace(0, x[-1], 100)
            fit_y = [cube_root_func(i, *vals[0]) for i in fit_x]

            plot_fit(x, y, cast(List[float], fit_x), fit_y, fit_to_zero)

        return float(fit_to_zero)

    def poly_exponential(
        self, x: List[float], y: List[float], _show_fit: bool, deg: int
    ) -> float:
        """Fit data to a poly-exponential, which is to say a function of the
        form :math:`a+e^{z}`, where :math:`z` is a polynomial.

        :param x: Noise scaling values.
        :type x: List[float]
        :param y: Expectation values.
        :type y: List[float]
        :param _show_fit: Plot data and resulting fitted function.
        :type _show_fit: bool
        :param deg: The degree of the polynomial in the exponential.
        :type deg: int
        :raises ValueError: Raised if the degree of the polynomial
            inputted is negative, or too high to fit to the data.
        :return: Extrapolation of data to the zero noise limit using the best fitting
            poly-exponential function of the specified degree.
        :rtype: float
        """

        # check that the degree of the polynomial is positive, and small enough to fit
        # to the inputted data
        if (deg + 2 > len(x)) or (deg < 0):
            raise ValueError(
                "The degree of the polynomial must be positive and can be at most is m-2, where m is the number of data points. In this case the largest permitted degree is %i."
                % (len(x) - 2)
            )

        # Fit data to polyexponential function
        vals = curve_fit(
            poly_exp_func, x, y, p0=[0, 1, *[-1 for i in range(deg)]], maxfev=10000
        )

        # Extrapolate function to zero noise limit
        fit_to_zero = poly_exp_func(0, *vals[0])

        # Plot data and fitted function
        if _show_fit:

            fit_x = np.linspace(0, x[-1], 100)
            fit_y = [poly_exp_func(i, *vals[0]) for i in fit_x]

            plot_fit(x, y, cast(List[float], fit_x), fit_y, fit_to_zero)

        return float(fit_to_zero)

    def exponential(
        self, x: List[float], y: List[float], _show_fit: bool, *args
    ) -> float:
        """Fit data to an exponential function. This is to say a function of the form :math:`a+e^{(b+x)}`.
        Note that this is a special case of the poly-exponential function.

        :param x: Noise scaling values.
        :type x: List[float]
        :param y: Expectation values.
        :type y: List[float]
        :param _show_fit: Plot data and resulting fitting function.
        :type _show_fit: bool
        :return: Extrapolation to zero noise limit using the best fitting exponential function.
        :rtype: float
        """

        # As the exponential function is a special case of the
        # poly-exponential function, it is called here
        return Fit.poly_exponential(self, x, y, _show_fit, 1)

    def polynomial(
        self, x: List[float], y: List[float], _show_fit: bool, deg: int
    ) -> float:
        """Fit data to a polynomial function.

        :param x: Noise scaling values.
        :type x: List[float]
        :param y: Expectation values.
        :type y: List[float]
        :param _show_fit: Plot data and resulting fitting function.
        :type _show_fit: bool
        :param deg: The degree of the function to fit to.
        :type deg: int
        :raises ValueError: Raised if the degree of the polynomial is negative,
            or too high to fit the data to.
        :return: Extrapolation to zero noise limit using the best fitting polynomial
            function of the specified degree.
        :rtype: float
        """

        # Raised if the degree of the polynomial requested is negative, or too high to fit the data to
        if (deg > len(x)) or (deg < 0):
            raise ValueError(
                "The degree of the polynomial must be positive and can be at most m-1, where m is the number of data points."
            )

        # Fit data to a polynomial
        fit = Polynomial.fit(x, [float(i) for i in y], deg=deg)

        # Extrapolate to the zero noise limit
        fit_to_zero = fit.convert().coef[0]

        # Plot data and fitted function
        if _show_fit:

            linspace = fit.linspace()
            fit_x = linspace[0]
            fit_y = linspace[1]

            plot_fit(x, y, fit_x, fit_y, fit_to_zero)

        return float(fit_to_zero)

    def linear(self, x: List[float], y: List[float], _show_fit: bool, *args) -> float:
        """Fit data to a linear function. This is to say a function of the form :math:`ax+b`.
        Note that this is a special case of the polynomial fitting function.

        :param x: Noise scaling values.
        :type x: List[float]
        :param y: Expectation values.
        :type y: List[float]
        :param _show_fit: Plot data and resulting fitted function.
        :type _show_fit: bool
        :return: Extrapolation to zero noise limit using the best fitting linear function.
        :rtype: float
        """
        # As this is a special case of a fit to a polynomial, the polynomial
        # fitting function is called here with a degree 1
        return Fit.polynomial(self, x, y, _show_fit, 1)

    def richardson(
        self, x: List[float], y: List[float], _show_fit: bool, *args
    ) -> float:
        """Use richardson extrapolation. This amounts to fitting to a polynomial of
        degree one less than the number of data points.

        :param x: Noise scaling values.
        :type x: List[float]
        :param y: Expectation values.
        :type y: List[float]
        :param _show_fit: Plot data and resulting fitted function.
        :type _show_fit: bool
        :return: Extrapolation to zero noise limit using Richardson extrapolation.
        :rtype: float
        """
        # As this is a special case of the polynomial fitting function, the polynomial fitting
        # function is called here with degree one less than the number of data points.
        return Fit.polynomial(self, x, y, _show_fit, len(x) - 1)


def plot_fit(
    x: List[float],
    y: List[float],
    fit_x: List[float],
    fit_y: List[float],
    fit_to_zero: float,
):
    """Plot expectation values at each noise level, and the fit to the data derived.

    :param x: Amounts by which the noise has been scaled
    :type x: List[float]
    :param y: Expectation values at each noise level
    :type y: List[float]
    :param fit_x: x coordinates at which to plot value of fitted function
    :type fit_x: List[float]
    :param fit_y: Value of fitted function at each noise scaling
    :type fit_y: List[float]
    :param fit_to_zero: The extrapolation of the fitted function to the zero noise limit
    :type fit_to_zero: float
    """

    fig = plt.figure()
    ax1 = fig.add_subplot(111)

    ax1.scatter(x, y, s=10, c="b", marker="s", label="data")
    ax1.plot(fit_x, fit_y, c="g", label="fit")

    ax1.scatter(0, fit_to_zero, s=10, c="r", marker="o", label="prediction")

    plt.legend()
    plt.show()


def digital_folding_task_gen(
    backend: Backend,
    noise_scaling: float,
    _folding_type: Folding,
    _allow_approx_fold: bool,
) -> MitTask:
    """
    Generates task transforming a circuit in order to amplify the noise. The noise
    is increased by a factor noise_scaling using the inputted folding method.

    :param backend: This will be used to compile the circuit after folding to ensure
        that the gate set matches those available on the backend.
    :type backend: Backend
    :param noise_scaling: The factor by which the noise is increased.
    :type noise_scaling: float
    :param _folding_type: The means by which the noise should be increased.
    :type _folding_type: Folding
    :param _allow_approx_fold:  Allows for the noise to be increased by an amount close to that requested, as
        opposed to by exactly the amount requested.
        This is necessary as there are cases where the exact noise scaling cannot be achieved.
        This occurs due to the discrete
        amounts by which the noise can be increased (i.e. the discrete amount by which one gate increases the noise).
    :type _allow_approx_fold: bool

    """

    def task(
        obj,
        mitex_wire: List[ObservableExperiment],
    ) -> Tuple[List[ObservableExperiment]]:
        """Increase the noise levels impacting the circuit by increasing the
        number of gates. This preserves the action of the circuit.

        :param mitex_wire: List of experiments
        :type mitex_wire: List[ObservableExperiment]
        :return: List of equivalent circuits, but with noise levels increased.
        :rtype: Tuple[List[ObservableExperiment]]
        """

        folded_circuits = []

        # For each circuit in the input wire, extract the circuit, apply the fold,
        # and perform the necessary compilation.
        for experiment in mitex_wire:

            # Apply the necessary folding method
            zne_circ = _folding_type(experiment.AnsatzCircuit.Circuit, noise_scaling, _allow_approx_fold)  # type: ignore

            # TODO: This additional compilation pass may result in the circuit noise being
            # increased too much, and should be removed or better accounted for.

            # This compilation pass was added to account for the case that
            # the inverse of a gate is not in the gateset of the backend.
            backend.default_compilation_pass(optimisation_level=0).apply(zne_circ)

            folded_circuits.append(
                ObservableExperiment(
                    AnsatzCircuit=AnsatzCircuit(
                        Circuit=zne_circ,
                        Shots=experiment.AnsatzCircuit.Shots,
                        SymbolsDict=experiment.AnsatzCircuit.SymbolsDict,
                    ),
                    ObservableTracker=experiment.ObservableTracker,
                )
            )

        return (folded_circuits,)

    return MitTask(_label="DigitalFolding", _n_in_wires=1, _n_out_wires=1, _method=task)


def extrapolation_task_gen(
    noise_scaling_list: List[float], _fit_type: Fit, _show_fit: bool, deg: int
) -> MitTask:
    """Generates task extrapolating to the zero noise limit using results from many folded circuits.

    :param noise_scaling_list: A list of the values by which the noise has been folded.
    :type noise_scaling_list: List[float]
    :param _fit_type: The function used to fit to the resulting data.
    :type _fit_type: Fit
    :param _show_fit: Plot data and resulting fitted function.
    :type _show_fit: bool
    :param deg: The degree of polynomials used.
    :type deg: int
    """

    def task(
        obj, base_exp_list: List[QubitPauliOperator], *args
    ) -> Tuple[List[QubitPauliOperator]]:
        """Returns expectation values corrected by extrapolation

        :param base_exp_list: List of expectation values corresponding to each experiment
        :type base_exp_list: List[QubitPauliOperator]
        :return: Each element of this tuple is a list of expectations for each
        experiment, all with noise scaled by a fixed amount.
        :rtype: Tuple[List[QubitPauliOperator]]
        """
        # Combine noise folding levels with the unfolded experiment
        all_fold_vals = [1, *noise_scaling_list]

        # Reformats to create list, where each list has fixed noise folding. Each element of
        # the list is a list of experiments
        expanded_args = [i for i in args]

        # Reformats so that all_fold_qpo_list_floats is a list with each element
        # corresponding to a fixed experiment. Each element itself is a dictionary {qps:list}
        # where the list is a list of expectation values from each noise folded experiment
        all_fold_qpo = [qpo for qpo in zip(base_exp_list, *expanded_args)]
        all_fold_qpo_list_floats = [
            {k: [d[k] for d in qpo_tuple] for k in qpo_tuple[0]._dict.keys()}
            for qpo_tuple in all_fold_qpo
        ]

        extrapolated = []

        # The list of expectations for each experiment is now used to
        # extrapolate to the ideal value
        for qpo_list_float in all_fold_qpo_list_floats:

            extrapolated.append(
                QubitPauliOperator(
                    {
                        qpo_k: _fit_type(  # type: ignore
                            obj, all_fold_vals, qpo_list_float[qpo_k], _show_fit, deg
                        )
                        for qpo_k in qpo_list_float
                    }
                )
            )

        return (extrapolated,)

    return MitTask(
        _label="CollateZNEResults",
        _n_in_wires=len(noise_scaling_list) + 1,
        _n_out_wires=1,
        _method=task,
    )


def gen_duplication_task(duplicates: int, **kwargs) -> MitTask:
    """Duplicate the inputted experiment wire

    :param duplicates: The number of times to duplicate the input wire.
    :type duplicates: int
    """

    def task(
        obj,
        mitex_wire: List[ObservableExperiment],
    ) -> Tuple[List[ObservableExperiment], ...]:
        """Duplicate the inputted experiment wire

        :param mitex_wire: List of experiments
        :type mitex_wire: List[ObservableExperiment]
        :raises ValueError: Raised if the number of duplications is less than 1
        :return: Many copies of the inputted wire
        :rtype: Tuple[List[ObservableExperiment]]
        """

        # Raise error if the number of duplications requested is less than 1
        if duplicates <= 0:
            raise ValueError(
                "The number of duplications must be greater than or equal to 1."
            )

        def copy_mitex_wire(wire: ObservableExperiment) -> ObservableExperiment:
            """Returns a single copy of the inputted wire

            :param wire: Pair of ansatz circuit and ObservableTracker
            :type wire: ObservableExperiment
            :return: single copy of inputted wire
            :rtype: ObservableExperiment
            """

            # Copy ansatz circuit
            new_ansatz_circuit = AnsatzCircuit(
                Circuit=wire.AnsatzCircuit.Circuit.copy(),
                Shots=copy.copy(wire.AnsatzCircuit.Shots),
                SymbolsDict=copy.copy(wire.AnsatzCircuit.SymbolsDict),
            )

            # copy qps and instantiate new measurement setup
            new_ot = ObservableTracker(
                QubitPauliOperator(
                    copy.copy(wire.ObservableTracker._qubit_pauli_operator._dict)
                )
            )
            new_wire = ObservableExperiment(
                AnsatzCircuit=new_ansatz_circuit, ObservableTracker=new_ot
            )
            return new_wire

        # Compose coppies into wire format
        if duplicates == 1:
            me_copy = [copy_mitex_wire(circuit_tuple) for circuit_tuple in mitex_wire]
            return (me_copy,)
        else:
            me_copies = [
                [copy_mitex_wire(circuit_tuple) for circuit_tuple in mitex_wire]
            ]
            for _ in range(duplicates - 1):
                mc = [copy_mitex_wire(circuit_tuple) for circuit_tuple in mitex_wire]
                me_copies.append(mc)
            return tuple(me_copies)

    return MitTask(
        _label=kwargs.get("_label", "Duplicate"),
        _n_out_wires=duplicates,
        _n_in_wires=1,
        _method=task,
    )


def gen_initial_compilation_task(
    backend: Backend, optimisation_level: int = 1
) -> MitTask:
    """Perform compilation before folding.

    :param backend: Backend to compile to
    :type backend: Backend
    :param optimisation_level: level of default compiler, defaults to 1
    :type optimisation_level: int, optional
    """

    def task(
        obj, wire: List[ObservableExperiment]
    ) -> Tuple[List[ObservableExperiment]]:
        """Performs initial compilation before folding. This is to ensure minimal compilation
        after folding, as this could disrupt by how much the noise is increased.

        :param wire: List of experiments
        :type wire: List[ObservableExperiment]
        :return: List of experiments compiled to run on the inputted backend
        :rtype: Tuple[List[ObservableExperiment]]
        """

        mapped_wire = []

        for obs_exp in wire:

            # Perform default compilation, tracking to which physical
            # qubits the initial qubits are mapped
            compiled_circ = obs_exp.AnsatzCircuit.Circuit.copy()
            cu = CompilationUnit(compiled_circ)
            backend.default_compilation_pass(
                optimisation_level=optimisation_level
            ).apply(cu)
            final_map = cu.final_map

            # TODO: This map needs to be used at the end of the
            # process in order to map the output pauli string back
            # to the pauli strings inputted by the user

            # Alter the qubit pauli operator so that it maps to the new physical qubits.
            orig_qpo_dict = obs_exp[1]._qubit_pauli_operator._dict.copy()
            new_qpo_dict = {}
            for orig_qps in orig_qpo_dict:
                orig_qps_dict = orig_qps.to_dict()
                new_qps_dict = {}
                for q in orig_qps_dict:
                    new_qps_dict[final_map[q]] = orig_qps_dict[q]
                new_qps = QubitPauliString(new_qps_dict)
                new_qpo_dict[new_qps] = orig_qpo_dict[orig_qps]
            new_qpo = QubitPauliOperator(new_qpo_dict)

            # Construct new list of experiments, but with compiled circuits.
            mapped_wire.append(
                ObservableExperiment(
                    AnsatzCircuit=AnsatzCircuit(
                        Circuit=cu.circuit,
                        Shots=obs_exp.AnsatzCircuit.Shots,
                        SymbolsDict=obs_exp.AnsatzCircuit.SymbolsDict,
                    ),
                    ObservableTracker=ObservableTracker(new_qpo),
                )
            )

        return (mapped_wire,)

    return MitTask(
        _label="CompileToBackend",
        _n_out_wires=1,
        _n_in_wires=1,
        _method=task,
    )


# TODO: Backend does not appear as input in documentation
def gen_ZNE_MitEx(backend: Backend, noise_scaling_list: List[float], **kwargs) -> MitEx:
    """Generates MitEx object which mitigates for noise using Zero Noise Extrapolation. This is the
    process by which noise is amplified incrementally, and the zero noise case arrived at by
    extrapolating backwards. For further explanantion see https://arxiv.org/abs/2005.10921.

    :param backend: Backend on which the circuits are to be run.
    :type backend: Backend
    :param noise_scaling_list: A list of the amounts by which the noise should be scaled.
    :type noise_scaling_list: List[float]
    :return: MitEx object performing noise mitigation by ZNE.
    :rtype: MitEx
    """
    _experiment_mitres = copy.copy(
        kwargs.get(
            "experiment_mitres",
            MitRes(backend),
        )
    )

    _experiment_mitex = copy.copy(
        kwargs.get(
            "experiment_mitex",
            MitEx(backend, _label="ExperimentMitex", mitres=_experiment_mitres),
        )
    )
    _experiment_taskgraph = TaskGraph().from_TaskGraph(_experiment_mitex)

    _optimisation_level = kwargs.get("optimisation_level", 0)
    _show_fit = kwargs.get("show_fit", False)
    _folding_type = copy.copy(kwargs.get("folding_type", Folding.circuit))
    _fit_type = copy.copy(kwargs.get("fit_type", Fit.linear))
    _deg = kwargs.get("deg", len(noise_scaling_list) - 1)
    _seed = kwargs.get("seed", None)
    _allow_approx_fold = kwargs.get("allow_approx_fold", True)

    np.random.seed(seed=_seed)

    for fold in noise_scaling_list:

        _label = str(fold) + "FoldMitEx"

        fold_mitex = copy.copy(
            kwargs.get(
                "fold_mitex", MitEx(backend, _label=_label, mitres=MitRes(backend))
            )
        )

        digital_folding_task = digital_folding_task_gen(
            backend, fold, _folding_type, _allow_approx_fold
        )
        fold_mitex.prepend(digital_folding_task)
        _experiment_taskgraph.parallel(fold_mitex)

    extrapolation_task = extrapolation_task_gen(
        noise_scaling_list, _fit_type, _show_fit, _deg
    )

    _experiment_taskgraph.append(extrapolation_task)
    _experiment_taskgraph.prepend(gen_duplication_task(len(noise_scaling_list) + 1))
    _experiment_taskgraph.prepend(
        gen_initial_compilation_task(backend, _optimisation_level)
    )
    return MitEx(backend).from_TaskGraph(_experiment_taskgraph)



# Copyright 2019-2021 Cambridge Quantum Computing
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import ABC, abstractmethod
from typing import List, Tuple, cast, Dict, Union

import numpy as np  #  type: ignore
import copy

from pytket.backends import Backend
from pytket.utils import QubitPauliOperator
from pytket.pauli import QubitPauliString  # type: ignore
from qermit import MitTask


class _BaseExCorrectModel(ABC):
    """
    Model for storing calibrations and correcting expectation values for some experiments.
    """

    @abstractmethod
    def calibrate(self, noisy_exp: List[float], exact_exp: List[float]) -> None:
        pass

    @abstractmethod
    def correct(self, noisy_expectation: float) -> float:
        return NotImplemented


class _PolyCDRCorrect(_BaseExCorrectModel):
    def __init__(self, degree: int, params: List[float] = []) -> None:
        super().__init__()
        if params:
            if len(params) != degree + 1:
                raise ValueError(
                    f"Degree {degree} polynomial requires {degree+1} parameters, "
                    f"{len(params)} were provided."
                )
        self.degree = degree
        self.params = params

    def calibrate(self, noisy_exp: List[float], exact_exp: List[float]) -> None:
        self.params = cast(
            List[float],
            cast(np.ndarray, np.polyfit(noisy_exp, exact_exp, self.degree)).tolist(),
        )

    def correct(self, noisy_expectation: float) -> float:
        return cast(
            float,
            sum(
                coef * (noisy_expectation ** power)
                for coef, power in zip(self.params, range(self.degree, -1, -1))
            ),
        )


def cdr_calibration_task_gen(backend: Backend, model: _BaseExCorrectModel) -> MitTask:
    """
    Uses calibration results from running characterisation circuits through a device
    and a noiseless simulator to characterise some model for correcting noisy expectation values.

    :param backend: Backend for storing characterisation model in.
    :type backend: Backend
    :param model: Model type to be calibrated and stored in backend.
    :type model: _BaseExCorrectModel
    """

    def cdr_calibration_task(
        obj,
        calibration_results: List[List[Tuple[QubitPauliOperator, QubitPauliOperator]]],
    ) -> Tuple[bool]:
        """
        For each experiments calibration results, calibrates a model of model type and stores it in backend.

        :param calibration_results: A list of noisy and noiseless expectation values for each
        experiment.
        :type calibration_results: List[List[Tuple[QubitPauliOperator, QubitPauliOperator]]]

        :return: A bool confirming characteriastion is complete
        :rtype: Tuple[bool]
        """
        counter = 0
        for calibration in calibration_results:
            # dict from QubitPauliString to Tuple[List[float], List[float]]
            # facilitates characteriastion of different QubitPauliStrings
            # for different experiments
            noisy_char_dict: Dict[QubitPauliString, List[float]] = dict()
            exact_char_dict: Dict[QubitPauliString, List[float]] = dict()

            for qpo_pair in calibration:
                noisy_qpo = qpo_pair[0]
                exact_qpo = qpo_pair[1]
                # go through strings in operator
                for key in noisy_qpo._dict:
                    # make sure keys are present (don't initialise at start incase indexing missing)
                    if key not in noisy_char_dict:
                        noisy_char_dict[key] = list()
                    if key not in exact_char_dict:
                        exact_char_dict[key] = list()
                    if key not in exact_qpo._dict:
                        raise ValueError(
                            "Given key in calibration task for Clifford Data Regression should be present in exact and noisy characterisation results."
                        )

                    noisy_char_dict[key].append(float(noisy_qpo._dict[key]))
                    exact_char_dict[key].append(float(exact_qpo._dict[key]))
            if backend.backend_info is None:
                raise ValueError("Backend has no backend_info attribute.")

            backend.backend_info.misc["CDR_" + str(counter)] = dict()
            # for each qubit pauli string in operator, add model for calibrating
            for key in noisy_char_dict:
                model.calibrate(noisy_char_dict[key], exact_char_dict[key])
                backend.backend_info.misc["CDR_" + str(counter)][key] = copy.copy(model)
            counter += 1

        return (True,)

    return MitTask(
        _label="CDRCalibrate",
        _n_in_wires=1,
        _n_out_wires=1,
        _method=cdr_calibration_task,
    )


def cdr_correction_task_gen(backend: Backend) -> MitTask:
    """
    For each QubitPauliOperator passed, corrects the given expectation for each
    internal QubitPauliString via some pre-characterised mode.

    :param backend: Backend object where characterisation is stored.
    :type backend: Backend
    """

    def cdr_correction_task(
        obj,
        noisy_expectation: List[QubitPauliOperator],
        calibration_complete: bool,
    ) -> Tuple[List[QubitPauliOperator]]:
        """
        :param noisy_expectation: QubitPauliOperator objects from some experiment
        which are to be corrected from some predefined characteriastion.
        :type noisy_expectation: List[QubitPauliOperator]
        :param calibration_complete: bool passed to method once calibration task has completed.
        :type calibration_complete: bool

        """
        if backend.backend_info is None:
            raise ValueError("Backend has no backend_info attribute.")

        corrected_expectations = []
        for i in range(len(noisy_expectation)):
            char_string = "CDR_" + str(i)
            if char_string not in backend.backend_info.misc:
                raise RuntimeError(
                    "CDR characterisation not stored in backend.charactersation attribute."
                )
            models = backend.backend_info.misc[char_string]
            new_qpo_dict = dict()
            for qps in noisy_expectation[i]._dict:
                new_qpo_dict[qps] = cast(
                    Union[int, float, complex],
                    models[qps].correct(float(noisy_expectation[i]._dict[qps])),
                )
            corrected_expectations.append(QubitPauliOperator(new_qpo_dict))

        return (corrected_expectations,)

    return MitTask(
        _label="CDRCorrect",
        _n_in_wires=2,
        _n_out_wires=1,
        _method=cdr_correction_task,
    )



# Copyright 2019-2021 Cambridge Quantum Computing
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List, Tuple, Dict
from .mitex import MitEx
from .mittask import MitTask, ObservableExperiment
from .utils import MeasurementCircuit

import copy

from pytket.backends import Backend
from pytket.transform import CXConfigType  # type: ignore
from pytket.partition import (  # type: ignore
    PauliPartitionStrat,
    measurement_reduction,
    GraphColourMethod,
)
from pytket import Bit
from pytket.pauli import QubitPauliString  # type: ignore


def measurement_reduction_task_gen(
    strategy: PauliPartitionStrat, method: GraphColourMethod, cx_config: CXConfigType
) -> MitTask:
    """
    Uses measurement reduction techniques available in the pytket.partition module to reduce
    the number of measurement circuits required to measure all expectations for an observable.

    :param strategy: Measurement reduction strategy used in measurement_reduction method
    :type strategy: PauliPartitionStrat
    :param method: Graph colouring method used in measurement_reduction method
    :type method:  GraphColourMethod
    :param cx_config: Configuration of CX gates for diagonlisation methods.
    :type cx_config: CXConfigType

    :return: MitTask object, taking and return same write type, but with added
        measurement circuits to ObservableTracker.
    :rtype: MitTask
    """

    def task(
        obj,
        measurement_wires: List[ObservableExperiment],
    ) -> Tuple[List[ObservableExperiment]]:
        new_wires = []
        # for each experiment
        for measurement_wire in measurement_wires:
            # circ is ansatz circuit which measurement circuits are appended to
            circ = measurement_wire[0][0]
            # symbols dict object
            symbols = measurement_wire[0][2]
            # ObservableTracker object for given experiment
            tracker = measurement_wire[1]
            # use measurement reduction method to get MeasurementSetup object
            # with reduced number of measurement circuits
            new_setup = measurement_reduction(
                list(tracker._qubit_pauli_operator._dict.keys()),
                strat=strategy,
                method=method,
                cx_config=cx_config,
            )
            # attach measurement configurations from MeasurementSetup to
            # copies of AnsatzCircuit
            # new measurement circuits
            measurement_circuits = []
            for measurement_circuit in new_setup.measurement_circs:
                full_circ = circ.copy()
                full_circ.append(measurement_circuit.copy())
                measurement_circuits.append(MeasurementCircuit(full_circ, symbols))
            # convert MeasurementBitMap objects to MeasurementInfo for ObservableTracker
            adder_info: Dict[
                int, List[Tuple[QubitPauliString, List[Bit], bool]]
            ] = dict()
            for i in range(len(measurement_circuits)):
                adder_info[i] = list()

            for qps, value in new_setup.results.items():
                for v in value:
                    circ_index = v.circ_index
                    bits = [Bit(i) for i in v.bits]
                    invert = v.invert
                    adder_info[circ_index].append((qps, bits, invert))

            for key in adder_info:
                tracker.add_measurement_circuit(
                    measurement_circuits[key], adder_info[key]
                )

            new_wires.append(
                ObservableExperiment(
                    AnsatzCircuit=measurement_wire[0], ObservableTracker=tracker
                )
            )
        return (new_wires,)

    return MitTask(
        _label="MeasurementReduction", _n_in_wires=1, _n_out_wires=1, _method=task
    )


def gen_MeasurementReduction_MitEx(backend: Backend, **kwargs) -> MitEx:
    """
    Returns a MitEx object that produces measurement circuits via Measurement Reduction
    methods.

    :param backend: Backend experiment is built around.
    :type backend: Backend
    :key mitex: MitEx object measurement reduction task is prepended to.

    :return: MitEx object performing Measurement reduction
    :rtype: MitEx
    """
    # mitex object to built measurement reduction
    _mitex = copy.copy(
        kwargs.get("mitex", MitEx(backend, _label="MeasurementReductionMitEx"))
    )
    _mitex._label = "MeasurementReductionMitEx"
    _strategy = kwargs.get("strategy", PauliPartitionStrat.NonConflictingSets)
    _method = kwargs.get("method", GraphColourMethod.Lazy)
    _cx_config = kwargs.get("cx_config", CXConfigType.Snake)
    # add measurement reduction task to start of method
    _mitex.prepend(measurement_reduction_task_gen(_strategy, _method, _cx_config))
    return _mitex



# Copyright 2019-2021 Cambridge Quantum Computing
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .mittask import (
    MitTask,
    IOTask,
    CircuitShots,
    Wire,
)
from typing import List, Tuple, Union, cast, Sequence
from .task_graph import TaskGraph
from pytket.backends import Backend, ResultHandle
from pytket.backends.backendresult import BackendResult
import networkx as nx  # type: ignore
import inspect
from copy import deepcopy


def backend_compile_circuit_shots_task_gen(
    backend: Backend, optimisation_level: int = 1
) -> MitTask:
    """
    For each circuit in passed List[CircuitShots] argument, pass
    circuit to compile_circuit method of given backend with given optimisation level.
    Returns new wire wherein each circuit has been compiled.

    :param backend: Backend object from which compile circuit method is called
    :type backend: Backend
    :param optimisation_level: Optimisation level of backend called.
    :type optimmisation_level: int
    """

    def task(obj, circ_shots: List[CircuitShots]) -> Tuple[List[CircuitShots]]:
        for cs in circ_shots:
            backend.compile_circuit(cs[0], optimisation_level=optimisation_level)
        return (circ_shots,)

    return MitTask(
        _label="CompileCircuitShots", _n_in_wires=1, _n_out_wires=1, _method=task
    )


def backend_handle_task_gen(backend: Backend) -> MitTask:
    """
    Passes every tuple of Circuit and Shots to the backend object MitTask
    is defined by, returning a handle for later retrieving results for each circuit.
    If different numbers of shots are passed, each circuit is run with the maximum number of shots.

    :param backend: Backend circuits are run through.
    :type backend: Backend
    :return: Pure function that adds passes circuits to backend and gets handles.
    :rtype: MitTask
    """

    def task(obj, circuit_wires: List[CircuitShots]) -> Tuple[List[ResultHandle]]:
        """
        :param circuit_wires: Circuits to be run on backend, number of shots to run of each.
        :type circuit_wires: List[CircuitShots]

        :return: ResultHandles from process_circuits method.
        :rtype: Tuple[List[ResultHandle]]
        """
        if len(circuit_wires) != 0:
            circs, shots = map(list, zip(*circuit_wires))

            return (backend.process_circuits(circs, n_shots=cast(Sequence[int], shots)),)
        else:
            return ([],)

    return MitTask(
        _label="CircuitsToHandles", _n_in_wires=1, _n_out_wires=1, _method=task
    )


def backend_res_task_gen(backend: Backend) -> MitTask:
    """
    For each ResultHandle passed to task, retrieves a BackendResult object from
    the backend the task is defined by.

    :param backend: backend holding results for handles.
    :type backend: Backend
    """

    def task(obj, handles: List[ResultHandle]) -> Tuple[List[BackendResult]]:
        return (backend.get_results(handles),)
        """
        :param handles: ResultHandle objects previously produced from backend.
        :type handles: List[ResultHandle]

        :return: For each ResultHandle in handles, a BackendResult object retrieved from backend.
        :rtype: Tuple[List[BackendResult]]
        """

    return MitTask(
        _label="HandlesToResults", _n_in_wires=1, _n_out_wires=1, _method=task
    )


class MitRes(TaskGraph):
    """
    A TaskGraph extension of mitigation of counts/shots for individual circuits.
    """

    def __init__(
        self,
        backend: Backend,
        _label: str = "MitRes",
    ) -> None:
        """
        MitRes objects are defined by the backend objects all circuits are executed on.


        :param backend: Pytket backend default constructor which tasks are generated from.
        :type backend: Backend
        :param label: Name for identification of MitRes object.
        :type label: str
        """
        # set member variables
        self._label = _label
        self.G = None

        # default constructor runs all circuits through passed Backend
        self._task_graph = nx.MultiDiGraph()

        c2h = backend_handle_task_gen(backend)
        h2r = backend_res_task_gen(backend)

        self._i, self._o = IOTask.Input, IOTask.Output
        self._task_graph.add_edge(self._i, c2h, key=(0, 0), data=None)
        self._task_graph.add_edge(c2h, h2r, key=(0, 0), data=None)
        self._task_graph.add_edge(h2r, self._o, key=(0, 0), data=None)

    def check_prepend_wires(self, task: Union[MitTask, "TaskGraph"]) -> bool:
        """
        Confirms that the number of out wires the passed task has is equal
        to the number of out wires from the input, and that the number
        of in wires the passed task has is 1. Also checks that the
        task.run attribute argument is List[CircuitShots] and that its
        return type is Tuple[List[CircuitShots]].

        :param task: MitTask or TaskGraph object for checking wire numbers of.
        :type task: Union[MitTask, Taskgraph]

        :return: True if task is suitable for prepending, False if not.
        :rtype: bool
        """
        sig = inspect.signature(task.run)
        params = list(sig.parameters.values())
        return (
            (task.n_out_wires == self.n_in_wires)
            and (task.n_in_wires == 1)
            and (len(params) == 1)
            and (params[0].annotation == List[CircuitShots])
            and (sig.return_annotation == Tuple[List[CircuitShots]])
        )

    def check_append_wires(self, task: Union[MitTask, "TaskGraph"]) -> bool:
        """
        Confirms that the number of in wires the passed task has is equal
        to the number of in wires to the output, and that the number
        of out wires the passed task has is 1. Also checks that the
        task.run attribute argument is List[BackendResult] and that its
        return type is Tuple[List[BackendResult]].

        :param task: MitTask or TaskGraph object for checking wire numbers of.
        :type task: Union[MitTask, Taskgraph]

        :return: True if task is suitable for apppending, False if not.
        :rtype: bool
        """
        sig = inspect.signature(task.run)
        params = list(sig.parameters.values())
        return (
            (task.n_in_wires == self.n_out_wires)
            and (task.n_out_wires == 1)
            and (len(params) == 1)
            and (params[0].annotation == List[BackendResult])
            and (sig.return_annotation == Tuple[List[BackendResult]])
        )

    def __str__(self) -> str:
        return f"<MitRes::{self._label}>"

    def __call__(self, circuits_wire: List[List[CircuitShots]]) -> Tuple[List[BackendResult]]:  # type: ignore[override]
        return cast(
            Tuple[List[BackendResult]], super().run(cast(List[Wire], circuits_wire))
        )

    def from_TaskGraph(self, task_graph: TaskGraph):
        """
        Returns a MitRes object from a TaskGraph object.

        :param task_graph: TaskGraph object to copy tasks from.
        :type task_graph: TaskGraph
        :return: Copied TaskGraph as MitRes
        :rtype: MitRes
        """
        if task_graph.n_in_wires != 1 or task_graph.n_out_wires != 1:
            raise TypeError(
                "Type signature of passed task_graph.run method does not equal MitRun.run type signature. Number of in and out wires does not match."
            )
        # can index as previous check means there should only be one edge
        input_parameters = list(
            inspect.signature(
                list(task_graph._task_graph.out_edges(task_graph._i))[0][1].run
            ).parameters.values()
        )
        if (
            len(input_parameters) != 1
            and input_parameters[0].annotation != List[CircuitShots]
        ):
            raise TypeError(
                "Type signature of passed task_graph.run method does not equal MitRes.run type signature. First MitTask in graph should expect a single argument of List[Circuitshots], but expects {}.".format(
                    input_parameters
                )
            )

        # can index as previous check means there should only be one edge
        return_annotation = inspect.signature(
            list(task_graph._task_graph.in_edges(task_graph._o))[0][0].run
        ).return_annotation
        if return_annotation != Tuple[List[BackendResult]]:
            raise TypeError(
                "Type signature of passed task_graph.run method does not equal MitRes.run type signature. Last MitTask in"
                "task graph should return Tuple[List[BackendResult]], but returns {}.".format(
                    return_annotation
                )
            )
        self._task_graph = deepcopy(task_graph._task_graph)
        self._label = task_graph._label
        return self

    def parallel(self, task: Union[MitTask, "TaskGraph"]):
        """
        Requests to add new MitTask/TaskGraph to TaskGraph object in parallel.
        Not permitted for MitRes, raises TypeError.

        :param task: New task to be added in parallel.
        :type task: MitTask
        """
        raise TypeError("MitRes.parallel forbidden.")

    def add_n_wires(self, num_wires: int):
        """
        Requests to add num_wires number of edges between the input vertex
        and output vertex, with no type restrictions. Not permitted for MitRes,
        raises TypeError.


        :param num_wires: Number of edges to add between input and output vertices.
        :type num_wires: int
        """
        raise TypeError("MitRes.add_n_wires forbidden.")

    def add_wire(self):
        """
        Requests to add a single edge between the input vertex and output vertex.
        Not permitted for MitRes, raises TypeError.
        """
        raise TypeError("MitRes.add_wire forbidden.")

    def run(self, circuit_shots: List[CircuitShots]) -> List[BackendResult]:  # type: ignore[override]
        """
        Overloaded run method from TaskGraph class to add type checking.
        A single experiment is defined by a Tuple containing a circuit to be run
        on some backend and the number of shots to take of said circuit.
        For each combination of Circuit and shots run returns a BackendResult object
        containing counts/shots.

        :param circuit_shotss: Each tuple in circuit_wires contains a Circuit to run on
            internal backends and the number of shots to take of said circuit.
        :type circuit_shots: List[CircuitShots]

        :return: A BackendResult object for each combination of circuit and shots.
        :rtype: List[BackendResult]
        """
        return self([circuit_shots])[0]



# Copyright 2019-2021 Cambridge Quantum Computing
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .mittask import (
    MitTask,
    IOTask,
    Wire,
)
import networkx as nx  # type: ignore
import graphviz as gv  # type: ignore
from typing import List, Union, Tuple, cast
import copy
from tempfile import NamedTemporaryFile


class TaskGraph:
    """
    The TaskGraph class stores a networkx graph where vertices
    are pure functions or tasks, and edges hold data.
    In the TaskGraph class these tasks and edges have no
    type restrictions, though for the run method to be succesful, the
    types of ports edges are attached to must match.

    :param _label: Name for identification of TaskGraph object.
    :type _label: str
    """

    def __init__(
        self,
        _label: str = "TaskGraph",
    ) -> None:
        # set member variables
        self._label = _label
        self.G = None

        # default constructor runs all circuits through passed Backend
        self._task_graph = nx.MultiDiGraph()

        self._i, self._o = IOTask.Input, IOTask.Output
        self._task_graph.add_edge(self._i, self._o, key=(0, 0), data=None)

    def from_TaskGraph(self, task_graph: "TaskGraph"):
        """
        Returns a new TaskGraph object from another TaskGraph object.

        :param task_graph: TaskGraph object to copy tasks from.
        :type task_graph: TaskGraph

        :return: Copied TaskGraph
        :rtype: TaskGraph
        """
        self._task_graph = copy.deepcopy(task_graph._task_graph)
        self._label = task_graph._label
        return self

    @property
    def tasks(self) -> List[MitTask]:
        """
        Returns a list of all tasks with both input and output ports
        in the TaskGraph.
        """
        return list(self._task_graph)[2:]

    def __call__(self, input_wires: List[Wire]) -> Tuple[List[Wire]]:
        return self.run(input_wires)

    @property
    def label(self) -> str:
        return self._label

    @property
    def n_in_wires(self):
        """
        The number of in wires to a TaskGraph object is defined as the number
        of out edges from the Input Vertex, as when called, a TaskGraph object
        calls the run method which stores input arguments as data on Input vertex
        output edges.
        """
        return len(self._task_graph.out_edges(self._i))

    @property
    def n_out_wires(self):
        """
        The number of out wires to a TaskGraph object is defined as the number
        of in edges to the Input Vertex, as when called, a TaskGraph object
        calls the run method which after running all tasks, returns
        the data on input edges to the Output Vertex as a tuple.
        """
        return len(self._task_graph.in_edges(self._o))

    def check_prepend_wires(self, task: Union[MitTask, "TaskGraph"]) -> bool:
        """
        Confirms that the number of out wires of the proposed task to prepend to the
        internal task_graph attribute matches the number of in wires to the graph.

        :param task: Wrapped pure function to prepend to graph
        :type task: Union[MitTask, "TaskGraph"]

        :return: True if prepend permitted
        :rtype: bool
        """
        return task.n_out_wires == self.n_in_wires

    def check_append_wires(self, task: Union[MitTask, "TaskGraph"]) -> bool:
        """
        Confirms that the number of in wires of the proposed task to append to the
        internal task_graph attribute matches the number of out wires to the graph.

        :param task: Wrapped pure function to append to graph
        :type task: Union[MitTask, "TaskGraph"]

        :return: True if append permitted
        :rtype: bool
        """
        return task.n_in_wires == self.n_out_wires

    def __str__(self):
        return f"<TaskGraph::{self._label}>"

    def __repr__(self):
        return str(self)

    def add_n_wires(self, num_wires: int):
        """
        Adds num_wires number of edges between the input vertex
        and output vertex, with no type restrictions.

        :param num_wires: Number of edges to add between input and output vertices.
        :type num_wires: int
        """
        for _ in range(num_wires):
            in_port = len(self._task_graph.out_edges(self._i, data=True))
            out_port = len(self._task_graph.in_edges(self._o, data=True))
            self._task_graph.add_edge(
                self._i, self._o, key=(in_port, out_port), data=None
            )

    def add_wire(self):
        """
        Adds a single edge between the input vertex and output vertex.
        """
        self.add_n_wires(1)

    # Add news task to start of TaskGraph
    def prepend(self, task: Union[MitTask, "TaskGraph"]):
        """
        Inserts new task to the start of TaskGraph._task_graph.
        All out edges from the Input vertex are wired as out edges from the task in the same port ordering (types must match).
        New edges also added from the Input vertex to the task (any type permitted), ports ordered in arguments order.

        :param task: New task to be prepended.
        :type task: MitTask
        """
        assert self.check_prepend_wires(task)
        task_copy = copy.copy(task)

        for i, edge in enumerate(list(self._task_graph.out_edges(self._i, keys=True))):
            self._task_graph.add_edge(
                task_copy, edge[1], key=(i, edge[2][1]), data=None
            )
            self._task_graph.remove_edge(edge[0], edge[1])

        for port in range(task_copy.n_in_wires):
            self._task_graph.add_edge(self._i, task_copy, key=(port, port), data=None)

    def append(self, task: Union[MitTask, "TaskGraph"]):
        """
        Inserts new task to end of TaskGraph._task_graph.
        All in edges to Output vertex are wired as in edges to task in same port ordering (types must match).
        New edges added from task to Output vertex (any type permitted), ports ordered in arguments order.

        :param task: New task to be appended.
        :type task: MitTask
        """
        assert self.check_append_wires(task)
        task_copy = copy.copy(task)
        for edge in list(self._task_graph.in_edges(self._o, keys=True)):
            self._task_graph.add_edge(edge[0], task_copy, key=edge[2], data=None)

            self._task_graph.remove_edge(edge[0], edge[1])

        for port in range(task_copy.n_out_wires):
            self._task_graph.add_edge(task_copy, self._o, key=(port, port), data=None)

    def decompose_TaskGraph_nodes(self):
        """
        For each node in self._task_graph, if node is a TaskGraph object, substitutes that node
        with the _task_graph structure held inside the node.
        """
        check_for_decompose = True
        while check_for_decompose == True:
            # get all nodes and iterate through them
            check_for_decompose = False
            node_list = list(nx.topological_sort(self._task_graph))
            for task in node_list:
                # => TaskGraph object with _task_graph attribute for decomposition
                if hasattr(task, "_task_graph"):
                    # relabel task names for ease of viewing wit visualisation methods
                    for sub_task in list(task._task_graph.nodes):
                        # in practice only IOTask
                        if hasattr(sub_task, "_label"):
                            sub_task._label = task._label + sub_task._label

                    task_in_edges = list(self._task_graph.in_edges(task, keys=True))
                    task_out_edges = list(self._task_graph.out_edges(task, keys=True))

                    task_input_out_edges = list(
                        task._task_graph.out_edges(task._i, keys=True)
                    )
                    task_output_in_edges = list(
                        task._task_graph.in_edges(task._o, keys=True)
                    )

                    if (
                        len(
                            set(task_input_out_edges).intersection(
                                set(task_output_in_edges)
                            )
                        )
                        > 0
                    ):
                        raise ValueError(
                            "Decomposition of TaskGraph node {}, not permitted: TaskGraph to be decomposed has edge between input and output vertices.".format(
                                task
                            )
                        )

                    # These two cases imply faulty TaskGraph construction
                    # Note that this check is only made as this is necessary constraint for decomposing TaskGraph nodes
                    # Faulty construction should be caughty at construction of TaskGraph object, including types
                    if len(task_in_edges) != len(task_input_out_edges):
                        raise TypeError(
                            "Decomposition of TaskGraph node {} not permitted: node "
                            "expects {} input wires but receives {}.".format(
                                task, len(task_input_out_edges), len(task_in_edges)
                            )
                        )
                    if len(task_out_edges) != len(task_output_in_edges):
                        raise TypeError(
                            "Decomposition of TaskGraph node {} not permitted: task_graph "
                            "expects {} output wires but node returns {}.".format(
                                task, len(task_input_out_edges), len(task_in_edges)
                            )
                        )

                    # remove all input and output edges from task._task_graph
                    # remove all input and output edges from self._task_graph task
                    # replace in_edges to task in self._task_graph with in_edges to first tasks in task
                    for outside_edge, inside_edge in zip(
                        task_in_edges, task_input_out_edges
                    ):
                        task._task_graph.remove_edge(inside_edge[0], inside_edge[1])
                        self._task_graph.remove_edge(outside_edge[0], outside_edge[1])
                        self._task_graph.add_edge(
                            outside_edge[0],
                            inside_edge[1],
                            key=(outside_edge[2][0], inside_edge[2][1]),
                            data=None,
                        )

                    for outside_edge, inside_edge in zip(
                        task_out_edges, task_output_in_edges
                    ):
                        task._task_graph.remove_edge(inside_edge[0], inside_edge[1])
                        self._task_graph.remove_edge(outside_edge[0], outside_edge[1])
                        self._task_graph.add_edge(
                            inside_edge[0],
                            outside_edge[1],
                            key=(inside_edge[2][0], outside_edge[2][1]),
                            data=None,
                        )

                    # add all remaining edges, filling the rest of the subsituted graph
                    self._task_graph.add_edges_from(task._task_graph.edges)
                    self._task_graph.remove_node(task)
                    check_for_decompose = True
                    break

    # eat, yum,
    def sandwich(
        self,
        prepend_task: Union[MitTask, "TaskGraph"],
        append_task: Union[MitTask, "TaskGraph"],
    ):
        """
        Does TaskGraph.prepend(prepend_task) and TaskGraph.append(append_task). Archaic but delicious.

        :param prepend_task: New task to be prepended.
        :type prepend_task: MitTask
        :param append_task: New task to be appended.
        :type append_task: MitTask
        """
        self.prepend(prepend_task)
        self.append(append_task)

    def parallel(self, task: Union[MitTask, "TaskGraph"]):
        """
        Adds new MitTask/TaskGraph to TaskGraph object in parallel. All task in edges wired as out edges from Input vertex. All task out_Edges wired as in edges to Output Vertex.

        :param task: New task to be added in parallel.
        :type task: MitTask
        """
        task = copy.copy(task)
        base_n_input_outs = len(self._task_graph.out_edges(self._i))
        for port in range(task.n_in_wires):
            self._task_graph.add_edge(
                self._i,
                task,
                key=(base_n_input_outs + port, port),
                data=None,
            )
        base_n_output_ins = len(self._task_graph.in_edges(self._o))
        for port in range(task.n_out_wires):
            self._task_graph.add_edge(
                task,
                self._o,
                key=(port, base_n_output_ins + port),
                data=None,
            )

    def run(self, input_wires: List[Wire]) -> Tuple[List[Wire]]:
        """
        Each task in TaskGraph is a pure function that produces output data
        from input data to some specification. Data is stored on edges of the
        internal _task_graph object.
        The run method first assigns each wire in the list to an output edge
        of the input vertex. If more wires are passed than there are edges, wire
        information is not assigned to some edge. If less wires are passed then there are edges,
        some edges will have no data and later computation will likely fail.
        Nodes (holding either MitTask or TaskGraph callable objects) on the graph undergo a topological sort to order them, and are then
        executed sequentially.
        To be executed, data from in edges to a task are passed arguments to the tasks
        _method, and data returned from method are assigned to out edges from a task.
        This process is repeated until all tasks are run, at which point all in edges
        to the output vertex wil have data on, with each data set returned in a tuple.

        :param input_wires: Each Wire holds information assigned as data to an output edge
            from the input vertex of the _task_graph.
        :type input_wires: List[Wire]

        :return: Data from input edges to output vertex, assigned as wires.
        :rtype: Tuple[List[Wire]]

        """
        for edge, wire in zip(
            self._task_graph.out_edges(self._i, data=True), input_wires
        ):
            edge[2]["data"] = wire

        # TODO Parallelism an option here, async an option for doing so.
        node_list = list(nx.topological_sort(self._task_graph))

        for task in node_list:
            if task in (self._i, self._o):
                continue
            in_edges = self._task_graph.in_edges(task, data=True, keys=True)
            inputs = [None] * len(in_edges)
            for _, _, ports, i_data in in_edges:
                assert i_data["data"] is not None
                inputs[ports[1]] = i_data["data"]
            out_edges = self._task_graph.out_edges(task, data=True, keys=True)

            outputs = task(inputs)
            assert len(out_edges) == len(outputs)
            for _, _, ports, o_data in out_edges:
                o_data["data"] = outputs[ports[0]]

        output_wire = [
            edge[2]["data"]
            for edge in list(self._task_graph.in_edges(self._o, data=True))
        ]
        return cast(Tuple[List[Wire]], tuple(output_wire))

    def get_task_graph(self) -> gv.Digraph:
        """
        Return a visual representation of the DAG as a graphviz object.

        :returns:   Representation of the DAG
        :rtype:     graphviz.DiGraph
        """
        G = gv.Digraph(
            "MEME",
            strict=True,
        )

        G.attr(rankdir="LR", ranksep="0.3", nodesep="0.15", margin="0")
        wire_color = "red"
        task_color = "darkolivegreen3"
        io_color = "green"
        out_color = "black"
        in_color = "white"

        boundary_node_attr = {"fontname": "Courier", "fontsize": "8"}
        boundary_nodes = {self._i, self._o}

        with G.subgraph(name="cluster_input") as c:
            c.attr(rank="source")
            c.node_attr.update(shape="point", color=io_color)
            for i in range(len(self._task_graph.out_edges(self._i))):
                c.node(
                    name=str(((str(self._i) + "out").replace("::", "_"), i)),
                    xlabel="Input" + str(i),
                    **boundary_node_attr,
                )

        with G.subgraph(name="cluster_output") as c:
            c.attr(rank="sink")
            c.node_attr.update(shape="point", color=io_color)
            for i in range(len(self._task_graph.in_edges(self._o))):
                c.node(
                    name=str(((str(self._o) + "in").replace("::", "_"), i)),
                    xlabel="Output",
                    **boundary_node_attr,
                )

        node_cluster_attr = {
            "style": "rounded, filled",
            "color": task_color,
            "fontname": "Times-Roman",
            "fontsize": "10",
            "margin": "5",
            "lheight": "100",
        }
        in_port_node_attr = {
            "color": in_color,
            "shape": "point",
            "weight": "2",
            "fontname": "Helvetica",
            "fontsize": "8",
            "rank": "source",
        }
        out_port_node_attr = {
            "color": out_color,
            "shape": "point",
            "weight": "2",
            "fontname": "Helvetica",
            "fontsize": "8",
            "rank": "sink",
        }
        count = 0
        for node, ndata in self._task_graph.nodes.items():
            if node not in boundary_nodes:
                with G.subgraph(name="cluster_" + node._label + str(count)) as c:
                    count = count + 1
                    c.attr(label=str(node._label), **node_cluster_attr)

                    n_in_ports = self._task_graph.in_degree(node)
                    if n_in_ports == 1:
                        c.node(
                            name=str(((str(node) + "in").replace("::", "-"), 0)),
                            **in_port_node_attr,
                        )
                    else:
                        for i in range(n_in_ports):
                            c.node(
                                name=str(((str(node) + "in").replace("::", "-"), i)),
                                xlabel=str(i),
                                **in_port_node_attr,
                            )

                    n_out_ports = self._task_graph.out_degree(node)
                    if n_out_ports == 1:
                        c.node(
                            name=str(((str(node) + "out").replace("::", "-"), 0)),
                            **out_port_node_attr,
                        )
                    else:
                        for i in range(n_out_ports):
                            c.node(
                                name=str(((str(node) + "out").replace("::", "-"), i)),
                                xlabel=str(i),
                                **out_port_node_attr,
                            )

        edge_attr = {
            "weight": "2",
            "arrowhead": "vee",
            "arrowsize": "0.2",
            "headclip": "true",
            "tailclip": "true",
        }
        for edge, edata in self._task_graph.edges.items():
            src_node, tgt_node, _ = edge
            src_port = edge[2][0]
            tgt_port = edge[2][1]
            src_nodename = str(((str(src_node) + "out").replace("::", "-"), src_port))
            tgt_nodename = str(((str(tgt_node) + "in").replace("::", "-"), tgt_port))
            G.edge(src_nodename, tgt_nodename, color=wire_color, **edge_attr)
        self.G = G
        return G

    def view_task_graph(self) -> None:
        """
        View the DAG.
        """
        G = self.get_task_graph()
        file = NamedTemporaryFile()
        G.view(file.name, quiet=True)



# Copyright 2019-2021 Cambridge Quantum Computing
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Counter, Tuple, List, Dict, Set, cast, Sequence
from pytket.circuit import Circuit, Qubit, Bit, CircBox  # type: ignore
from pytket.passes import DecomposeBoxes, FlattenRegisters  # type: ignore
from qermit import MitTask, CircuitShots
import itertools
from collections import namedtuple
from pytket.backends import Backend
from pytket.backends.backendresult import BackendResult
from pytket.utils.outcomearray import OutcomeArray
import numpy as np
from math import ceil
import networkx as nx  # type: ignore


PartialCorrelatedNoiseCharacterisation = namedtuple(
    "PartialCorrelatedNoiseCharacterisation",
    ["Distance", "CorrelatedEdges", "CorrelationToMatrix"],
)


def gen_state_polarisation_dicts(correlations: nx.Graph) -> List[Dict]:
    """Generating all four basis states for capturing two-qubit correlations requires circuits
    producing the following states:
    A B
    0 0
    0 1
    1 0
    1 1
    Correlations between two-qubits can be assumed captured for a given process if tomography circuits
    are run preparing these basis states. A and B also function as two patterns for initialisating qubits;
    if 4 circuits are produced for Qubit 1 with pattern A, and the same circuits are extended to
    initialise some Qubits 2 and 3 in pattern B, then when run for some process, the given circuits
    will capture all correlations betweeen Qubits 1 and 2 and Qubits 1 and 3, and only half (00 & 11)
    for Qubits 2 and 3. If correlations between 2 and 3 need to be captured, a new set of circuits must be run
    with 2 in pattern A and 3 in pattern B (or vice versa). We can summarise this information with the following
    rows as:
    1 2 3
    A B B
    2 3
    A B
    Translated to basis state preparation, this would require 8 circuits, 4 for each row (though n.b. that with
    later optimisation this can be reduced to 6, as the second row only needs to capture 01 and 10 correlations for
    Qubits 2 and 3).
    For any Graph, where nodes are qubits and edges capture desired correlations between qubits, this algorithm
    finds Qubit & pattern combinations covering all correlations with a number of rows equal to the
    number of nodes in the largest fully connected subgraph - 1 (n.b. each row then corresponds to 4 basis states
    meaning 4*(number of nodes in the largest fully connected subgraph - 1) circuits are required to characterise
    correlated noise)(also n.b. this is with the assumption that strong correlations between adjacent qubits
    does not imply weak correlations between said adjacent qubits)

    :param correlations: Graph object wherein adjacent nodes have two qubit correlations
    :type correlations: nx.Graph

    :return: A list of dictionaries, each dictionary from node to basis state preparation pattern
    :rtype: List[Dict[UnitID, bool]]
    """
    # dicts for holding polarisation pattern
    state_polarisation_dicts: List[Dict] = []
    used_nodes: Set[Qubit] = set()
    for node in sorted(set(correlations.nodes)):
        # also prevents case where neighbour==node causing issue (some hardware graphs are defined in this way...)
        used_nodes.add(node)
        for neighbour in sorted(set(correlations.neighbors(node))):
            # corelations are bidirectional as is graph
            # check prevents correlations being added twice
            if neighbour not in used_nodes:
                value_set = False
                # note state_polarisation_dicts is a list, longest polarisation will always be first entry
                for states in state_polarisation_dicts:
                    # break to not add correlations to multiple states! (though this will happen naturally in some cases and is fine)
                    if node not in states:
                        # node not in states => neighbour not in states
                        # as it implies neighbour has not been node yet
                        states[node] = True
                        states[neighbour] = False
                        value_set = True
                        break
                    if node in states and neighbour not in states:
                        # set to opposite polarisation
                        states[neighbour] = not states[node]
                        value_set = True
                        break
                    if node in states and neighbour in states:
                        if states[node] != states[neighbour]:
                            # correlation already captured
                            value_set = True
                            break
                if not value_set:
                    # correlation can't be captured with any existing state, prep a new one
                    state_polarisation_dicts.append(
                        dict([(node, True), (neighbour, False)])
                    )
    # Sort as number of circuits run total can be minimised further later with help of reverse ordering
    state_polarisation_dicts.sort(key=len)
    return state_polarisation_dicts


def gen_partial_tomography_circuits(
    state_polarisations: List[Dict], x_prep_box: CircBox
) -> Tuple[List[Circuit], Dict]:
    """Basis states for correlations between two qubits can be prepared with the following patterns:
    A B
    0 0
    0 1
    1 0
    1 1
    state_polarisations gives a List[Dict], for which each dictionary is from UnitID to pattern.
    gen_partial_tomography_circuits converts these patterns to circuits, with basis state prep given by x_prep_bo (i.e.
    gate primitive for X for desired hardware). Returned Dict has tuple((UnitID, UnitID),(int, int)) as keys and int
    as values, instructing later characterisation procedures to which BackendResult are required to capture some
    correlation between two qubits.

    :param state_polarisations: Guides basis state construction, as given by gen_state_polarisation_dicts
    :type state_polarisations: List[Dict]
    :param x_prep_box: Single qubit CircBox giving Pyktet Circuit for preparing |1> state for a single qubit
    :type x_prep_box: CircBox

    :return: Characterisation circuits, and dict for finding results for correlation basis and Qubits
    :rtype: Tuple(Tuple(UnitID, UnitID), Tuple(int, int))
    """
    all_circuits: List[Circuit] = []
    correlations_dict = dict()
    for states in state_polarisations:
        # for two qubits there are 4 possible basis states
        basis_state_circuits = [Circuit(), Circuit(), Circuit(), Circuit()]
        # for later organsiation all correlations a given 'state polarisation; captures
        true_keys: Set[Qubit] = set()
        false_keys: Set[Qubit] = set()
        # track measurements as they are added
        measure_dict: Dict[Qubit, Bit] = dict()

        for node in states:
            # add qubit to each basis state circuit
            for c in basis_state_circuits:
                c.add_qubit(node)
            # if true, implies qubit needs to be in 0 state for first two basis states, and 1 state for second two
            if states[node]:
                true_keys.add(node)
                basis_state_circuits[2].add_circbox(x_prep_box, [node])
                basis_state_circuits[3].add_circbox(x_prep_box, [node])
            # if false, implies qubit needs to be in 0 state for first and third basis states, and 1 state for second and last
            else:
                false_keys.add(node)
                basis_state_circuits[1].add_circbox(x_prep_box, [node])
                basis_state_circuits[3].add_circbox(x_prep_box, [node])
            # add entry to measurement dict, measure qubits later
            measure_dict[node] = Bit(len(measure_dict))

        # make circuit proper for running -> decompose circuit box, add physical barrier to prevent further optimisation
        for c in basis_state_circuits:
            DecomposeBoxes().apply(c)
            c.add_barrier(c.qubits)
        # add measures to circuit
        for k in measure_dict:
            v = measure_dict[k]
            for c in basis_state_circuits:
                c.add_bit(v)
                c.Measure(k, v)

        # fill correlations dict
        base_len = len(all_circuits)
        for t in sorted(true_keys):
            for f in sorted(false_keys):
                # keys a pair of circuit qubits pertaining to some correlation
                # values are Index in results, state measured and classical bits required to retrieve correlation result
                correlations_dict[(t, f)] = [
                    (base_len, (0, 0), (measure_dict[t], measure_dict[f])),
                    (base_len + 1, (0, 1), (measure_dict[t], measure_dict[f])),
                    (base_len + 2, (1, 0), (measure_dict[t], measure_dict[f])),
                    (base_len + 3, (1, 1), (measure_dict[t], measure_dict[f])),
                ]
        # add circuits to all circuits
        for c in basis_state_circuits:
            all_circuits.append(c)

    return (all_circuits, correlations_dict)


def partial_spam_setup_task_gen(backend: Backend, correlation_distance: int) -> MitTask:
    """
    Sets up required information for characterising and correcting SPAM noise.
    Includes check for whether SPAM Characterisation needs to be run.

    :param backend: Default Backend characterisation and experiment are executed on.
    :type backend: Backend
    :param correlations_distance: Distance over Backend Connectivity graph over which correlations in Qubit SPAM Noise is expected.
    :type correlations_distance: int

    :return: A MitTask object that completes said task.
    :rtype: MitTask
    """

    def task(
        obj, circ_shots: List[CircuitShots]
    ) -> Tuple[List[CircuitShots], List[Dict[Qubit, Bit]], bool]:
        characterisation = True

        if correlation_distance < 1:
            raise ValueError(
                "Correlations distance must be 1 or greater for characterisation of partial spam correlations."
            )
        if backend.backend_info is None:
            raise ValueError("Backend has no characterisation attribute.")
        else:
            if "CorrelatedSpamCorrection" in backend.backend_info.misc:
                # check correlations distance
                if (
                    backend.backend_info.misc["CorrelatedSpamCorrection"].Distance
                    is correlation_distance
                ):
                    characterisation = False
        qubit_bit_maps = [c[0].qubit_to_bit_map for c in circ_shots]
        return (circ_shots, qubit_bit_maps, characterisation)

    return MitTask(_label="SPAMSetup", _n_in_wires=1, _n_out_wires=3, _method=task)


def partial_correlated_spam_circuits_task_gen(
    backend: Backend, calibration_shots: int, correlation_distance: int
) -> MitTask:
    """
    For Given Backend, returns circuits characterisating SPAM noise for up to correlation_distance distance on the Backend connectivity graph.

    :param backend: Default Backend characterisation and experiment are executed on.
    :type backend: Backend
    :param calibration_shots: Number of shots required for each characterisation circuit
    :type calibration_shots: int
    :param correlations_distance: Distance over Backend Connectivity graph over which correlations in Qubit SPAM Noise is expected.
    :type correlations_distance: int

    :return: A MitTask object that produces Characterisation circuits.
    :rtype: MitTask
    """

    def task(obj, characterise: bool) -> Tuple[List[CircuitShots], Dict]:
        """
        Pure Function that transforms experimental circuit information to characterisation circuits and dict
        for storing correlations for SPAM noise.

        :param characterise: True if characterisation required
        :type characterise: bool

        :return: A list of Characterisation circuits and dict from correlation to index in Characterisation Circuits list
        and classical bits required derive characterisation for correlation.
        :rtype: Tuple[List[CircuitShots], Dict]

        """
        if characterise:
            # get basic graph, representing nearest neighbour correlations
            correlations = nx.Graph()

            if backend.backend_info is None:
                raise ValueError("Backend has no backend_info attribute.")
            correlations.add_edges_from(backend.backend_info.architecture.coupling)
            # to allow non-nearest neighbour correlations, add edges to graph
            # to reflect larger distance correlations
            for _ in itertools.repeat(None, correlation_distance - 1):
                new_correlations = []
                for node in sorted(set(correlations.nodes)):
                    # Forced to spell this way by itertools...
                    neighbors = sorted(set(correlations.neighbors(node)))
                    # NoiseModel Backends and some Hardware Backends Will have
                    # Couplings specified between a Node to itself
                    if node in neighbors:
                        neighbors.remove(node)
                    for pair in itertools.combinations(neighbors, r=2):
                        new_correlations.append(pair)
                correlations.add_edges_from(new_correlations)

            # gen basis state prep polarisations required to cover all correlations
            sp_dicts = gen_state_polarisation_dicts(correlations)

            # produce basis state preparation circuit
            xcirc = Circuit(1).X(0)
            backend.compile_circuit(xcirc)
            FlattenRegisters().apply(xcirc)
            x_prep_box = CircBox(xcirc)

            # get tomography circuits
            circs_dict = gen_partial_tomography_circuits(sp_dicts, x_prep_box)
            # translate circuits to CircShots type required by wire
            circ_shots = [
                CircuitShots(Circuit=c, Shots=calibration_shots) for c in circs_dict[0]
            ]
            return (circ_shots, circs_dict[1])
        else:
            return ([], dict())

    return MitTask(
        _label="SPAMCalibrationCircuits", _n_in_wires=1, _n_out_wires=2, _method=task
    )


def order_counts(counts: Counter[Tuple[int, ...]]) -> List[int]:
    """
    Helper method for organising two-qubit correlations.
    :param counts: Counter object as returned by BackendResult, giving two qubit counts for desired correlation
    :type counts: Counter[Tuple[int, ...]]

    :return: A four element list, giving counts for the (0,0), (0,1), (1,0) and (1,1) states in order.
    :rtype: List[int]
    """
    ordered_counts = [0, 0, 0, 0]
    if (0, 0) in counts:
        ordered_counts[0] = counts[(0, 0)]
    if (0, 1) in counts:
        ordered_counts[1] = counts[(0, 1)]
    if (1, 0) in counts:
        ordered_counts[2] = counts[(1, 0)]
    if (1, 1) in counts:
        ordered_counts[3] = counts[(1, 1)]
    return ordered_counts


def characterise_correlated_spam_task_gen(
    backend: Backend, correlations_distance: int, calibration_shots: int
) -> MitTask:
    """
    Produces SPAM characterisation matrices for each correlation expected.

    :param backend: Backend which characterisation to be stored in
    :type backend: Backend
    :param correlations_distance: Correlation Distance which charactersation was produced for - only required for setting in backend.
    :type correlations_distance: int

    :return: A MitTask object that assigns a PartialCorrelatedNoiseCharacterisation to backend.
    :rtype: MitTask

    """

    def task(obj, results: List[BackendResult], correlations_dict: Dict) -> Tuple[bool]:
        """
        Pure function that transforms BackendResult objects to a PartialCorrelatedNoiseCharacterisation that is stored
        in backend, only returning a bool for confirming this task has completed.

        :param results: Results retrieved from defined backend.
        :type results: List[BackendResult]
        :param correlations_dict: For a given graph edge, gives the index of results for which results for a given correlation ((0,0) or (0,1) etc) can be found from given bits.
        :type correlations_dict: dict
        """

        # if condition not met, nothing is run as characterisation is not required
        # guarantees characterisation isn't written over
        if len(results) > 0 and len(correlations_dict) > 0:
            correlation_to_mat = dict()
            correlated_edges = []
            for nodes in correlations_dict:
                value = correlations_dict[nodes]
                cal_mat = np.empty((4, 4))
                correlated_edges.append(nodes)
                # 00->0; 01->1; 10->2; 11->3; measured->row; prepared->column
                cal_mat[:, 0] = order_counts(
                    results[value[0][0]].get_counts(list(value[0][2]))
                )
                cal_mat[:, 1] = order_counts(
                    results[value[1][0]].get_counts(list(value[1][2]))
                )
                cal_mat[:, 2] = order_counts(
                    results[value[2][0]].get_counts(list(value[2][2]))
                )
                cal_mat[:, 3] = order_counts(
                    results[value[3][0]].get_counts(list(value[3][2]))
                )
                # normalised matrix
                cal_mat = cal_mat / calibration_shots
                v = np.linalg.inv(cal_mat)
                v[v < 0] = 0
                v /= sum(v)
                correlation_to_mat[nodes] = v

            if backend.backend_info is None:
                raise ValueError("Backend has no backend_info attribute.")
            backend.backend_info.add_misc(
                "CorrelatedSpamCorrection",
                PartialCorrelatedNoiseCharacterisation(
                    correlations_distance, correlated_edges, correlation_to_mat
                ),
            )
        return (True,)

    return MitTask(
        _label="SPAMCharacterisation", _n_in_wires=2, _n_out_wires=1, _method=task
    )


def collate_corrected_counts(
    width: int,
    c_bits: Dict[Bit, int],
    correlation_counts: List[Tuple[Tuple[Bit, Bit], List[int], int]],
    total_counts: int,
) -> BackendResult:
    """
    Converts a list of counts for different pairs of indices into a single BackendResult object. Width must be
    greater than largest index (assume 0 indexing).

    :param width: Number of bits in output result
    :type width: int
    :param corrected_counts: Each entry gives counts for a different set of four basis states for the given Bits.
    :type corrected_counts: List[Tuple(Tuple(Bit, Bit), List[int], int)]

    :return: A BackendResult object capturing all these counts
    :rtype: BackendResult
    """
    collated_counts = dict.fromkeys(itertools.product([0, 1], repeat=width), float(1))
    # For each correlation, retrieve bits corrected counts are for
    for correlation in correlation_counts:
        indices = [c_bits[correlation[0][0]], c_bits[correlation[0][1]]]
        # division acts as normalisation, retrieve counts
        counts = [x / correlation[2] for x in correlation[1]]
        # update dictionary containing all counts
        # due to symmetry of binary strings, each string will have the same number of counts added for each correlations

        # need to find allow states from given, and proportion
        # i.e. if n shots taken, find proportion of given state and return thins info
        # possible states?
        for key in collated_counts:
            # if else statements for assigning corrected counts to binary representing given count
            if key[indices[0]] == 0:
                if key[indices[1]] == 0:
                    collated_counts[key] = counts[0] * collated_counts[key]
                else:
                    collated_counts[key] = counts[1] * collated_counts[key]
            elif (
                key[indices[1]] == 0
            ):  # not prior if => key[indices[0]] == 1, so just check second bit though maybe safer to check...
                collated_counts[key] = counts[2] * collated_counts[key]
            else:  # every other option exhausted, so must be this
                collated_counts[key] = counts[3] * collated_counts[key]

    # convert collated counts to a BackendResult object
    normalisation_factor = sum(collated_counts.values())
    # use first shots value in list i.e. correlation_counts[0][2]
    for k in collated_counts:
        collated_counts[k] = (collated_counts[k] / normalisation_factor) * total_counts
    counter = Counter(
        {
            OutcomeArray.from_readouts([key]): ceil(val)
            for key, val in collated_counts.items()
        }
    )
    return BackendResult(counts=counter, c_bits=cast(Sequence[Bit], c_bits))


def correct_partial_correlated_spam_task_gen(
    backend: Backend,
) -> MitTask:
    """
    Produces and returns a MitTask object to apply SPAM correction to BackendResult objects.
    :param backend: Backend from which Spam Correction characterisation is retrieved.
    :type backend: Backend
    """

    def task(
        obj,
        results: List[BackendResult],
        qubit_readouts: List[Dict],
        characterised: bool,
    ) -> Tuple[List[BackendResult]]:

        if backend.backend_info is None:
            raise ValueError("Backend has no backend_info attribute.")
        if "CorrelatedSpamCorrection" in backend.backend_info.misc:
            cnc = backend.backend_info.misc["CorrelatedSpamCorrection"]
        else:
            raise ValueError(
                "'CorrelatedSpamCorrection' not characterised for Backend."
            )
        output_results = []
        for result, qubit_readout in zip(results, qubit_readouts):
            desired_qubits = qubit_readout.keys()
            single_correlation_corrections = []
            total_counts = len(result.get_shots())
            for edge in cnc.CorrelatedEdges:
                # only get correlated qubit edge if actually measured
                # if not checked, wouldn't be able to retrieve bits for qubits not present
                if edge[0] in desired_qubits and edge[1] in desired_qubits:
                    # edge in terms of Qubit, find bit as provided by experiment Circuit
                    bits = (qubit_readout[edge[0]], qubit_readout[edge[1]])
                    # get correction matrix from Backend characterisation (already inverted)
                    mat = cnc.CorrelationToMatrix[edge]
                    # order counts converts key -> value counts with a list where 0th index gives 00 result, 1st 01, 2nd 10 and 3rd 11.
                    edge_counts = order_counts(result.get_counts(list(bits)))
                    # correct this list for SPAM with the inverted correction matrix
                    corrected_counts = mat.dot(edge_counts)
                    # add corrected counts for given correlation in terms of bits results represent, along with total shots
                    single_correlation_corrections.append(
                        (bits, corrected_counts, sum(edge_counts))
                    )
            # convert list of 2 qubit counts for different correlated bits into a single BackendResult

            corrected_result = collate_corrected_counts(
                len(result.c_bits),
                result.c_bits,
                single_correlation_corrections,
                total_counts,
            )

            output_results.append(corrected_result)

        return (output_results,)

    return MitTask(_label="SPAMCorrection", _n_in_wires=3, _n_out_wires=1, _method=task)



# Copyright 2019-2021 Cambridge Quantum Computing
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List, Tuple, Counter, cast, Sequence
from functools import reduce
import operator

from qermit import (
    MitRes,
    MitTask,
    CircuitShots,
)
from qermit.taskgraph.mitex import backend_compile_circuit_shots_task_gen
from pytket import Circuit, Bit
from pytket.transform import Transform  # type: ignore
import copy
from math import ceil
from pytket.backends import Backend
from pytket.backends.backendresult import BackendResult
from pytket.utils.outcomearray import OutcomeArray
from pytket.tailoring import UniversalFrameRandomisation, PauliFrameRandomisation  # type: ignore
from pytket.passes import RebaseUFR  # type: ignore
from enum import Enum


class FrameRandomisation(Enum):
    def PauliFrameRandomisation(
        circuit: Circuit, shots: int, samples: int
    ) -> List[CircuitShots]:
        """
        Uses the pytket.tailoring method PauliFrameRandomisation  to return a list of
        circuits corresponding to instances of frame randomisation for the input circuit.

        :param circuit: Circuit to have frame randomisation applied.
        :type circuit: Circuit
        :param shots: Number of shots of each frame randomisation circuit to take.
        :type shots: int
        :param samples: Number of frame randomisation instances to sample.
        :type samples: int
        """
        pfr = PauliFrameRandomisation()
        pfr_shots = ceil(shots / samples)
        RebaseUFR().apply(circuit)
        Transform.RebaseToCliffordSingles().apply(circuit)
        pfr_circuits = pfr.sample_circuits(circuit, samples)
        return [CircuitShots(Circuit=c, Shots=pfr_shots) for c in pfr_circuits]

    def UniversalFrameRandomisation(
        circuit: Circuit, shots: int, samples: int
    ) -> List[CircuitShots]:
        """
        Uses the pytket.tailoring method UniversalFrameRandomisation  to return a list of
        circuits corresponding to instances of frame randomisation for the input circuit.
        UniversalFrameRandomisation rebases the circuit to a gate set such that
        the whole circuit is a single cycle for frame randomisation. This is possible with
        the additional noise assumption on top of the regular Frame Randomisation assumptions
        that Rz(-x) and Rz(x) incur similar noise for any angle x.

        :param circuit: Circuit to have frame randomisation applied.
        :type circuit: Circuit
        :param shots: Number of shots of each frame randomisation circuit to take.
        :type shots: int
        :param samples: Number of frame randomisation instances to sample.
        :type samples: int
        """
        ufr = UniversalFrameRandomisation()
        ufr_shots = ceil(shots / samples)
        RebaseUFR().apply(circuit)
        ufr_circuits = ufr.sample_circuits(circuit, samples)
        return [CircuitShots(Circuit=c, Shots=ufr_shots) for c in ufr_circuits]


def frame_randomisation_circuits_task_gen(
    samples: int, _fr_type: FrameRandomisation
) -> MitTask:
    """
    Returns a MitTask object that produces Frame Randomisation circuits for some wire of experiment circuits.

    :param samples: Number of samples of frame randomisation circuits to take for each circuit in wire.
    :type samples: int
    """

    def task(obj, circs_shots: List[CircuitShots]) -> Tuple[List[CircuitShots]]:
        """
        :param circ_shots: A list of tuple of circuits and shots. Each circuit has frame randomisation applied.
        The number of shots of each frame randomisation circuit is ceil(shots/samples)
        :type circ_shots: List[CircuiShots]

        :return: Frame Randomisation circutis
        :rtype: Tuple[List[CircuitShots]]
        """

        all_fr_circs_shots = []
        for circ, shots in circs_shots:
            fr_circshots = _fr_type(circ, shots, samples)  # type: ignore
            # frame randomisation type not callable as no default enum
            all_fr_circs_shots.extend(fr_circshots)
        return (all_fr_circs_shots,)

    return MitTask(
        _label="GenerateFrameRandomisationCircuits",
        _n_in_wires=1,
        _n_out_wires=1,
        _method=task,
    )


def frame_randomisation_result_task_gen(samples: int) -> MitTask:
    """
    Returns a MitTask object that sequentially collates samples number of BackendResult objects into single
    BackendResult object. These collated BackendResult objects should include all frame experiments for a single
    original circuit.

    :param samples: Number of frame randomisation instances created in the first place, used to suitably
        collate results.
    :type samples: int
    """

    def task(
        obj,
        all_fr_results: List[BackendResult],
    ) -> Tuple[List[BackendResult]]:
        """
        :param all_fr_results: A list of BackendResult objects for all frame randomisations for all experiment circuits.
        :type: all_fr_results: List[BackendResult]

        :return: Collated BackendResult objects, reflecting frame randomisation procedure
        :rtype: Tuple[List[BackendResult]]
        """
        chunked_results = [
            all_fr_results[i : i + samples]
            for i in range(0, len(all_fr_results), samples)
        ]
        results = []
        for chunk in chunked_results:
            all_counts = [result.get_counts() for result in chunk]
            combined_counts = reduce(operator.add, all_counts)
            outcome_array = {
                OutcomeArray.from_readouts([key]): val
                for key, val in combined_counts.items()
            }
            outcome_counts = Counter(outcome_array)
            results.append(
                BackendResult(
                    counts=outcome_counts, c_bits=cast(Sequence[Bit], chunk[0].c_bits)
                )
            )
        return (results,)

    return MitTask(
        _label="CollateFrameRandomisationResults",
        _n_in_wires=1,
        _n_out_wires=1,
        _method=task,
    )


def gen_Frame_Randomisation_MitRes(backend: Backend, samples: int, **kwargs) -> MitRes:
    """
    Produces a MitRes object that applies FrameRandomisation techniques to experiment circuits.

    :param backend: Backend which experiments are default run through.
    :type backend: Backend
    :param samples: Number of Frame Randomisation instances sampled in Frame Randomisation.
    :type samples: int

    :key mitres: MitRes object FrameRandomisation MitRes built around if given.
    :key frame_randomisation: FrameRandomisation Enum passed to specify method used.
        Default set to FrameRandomisation.UniversalFrameRandomisation.
    """
    _mitres = copy.copy(
        kwargs.get("mitres", MitRes(backend, _label="FrameRandomisationMitRes"))
    )
    _fr_type = copy.copy(
        kwargs.get(
            "frame_randomisation", FrameRandomisation.UniversalFrameRandomisation
        )
    )
    _mitres.prepend(backend_compile_circuit_shots_task_gen(backend))
    _mitres.sandwich(
        frame_randomisation_circuits_task_gen(samples, _fr_type),
        frame_randomisation_result_task_gen(samples),
    )
    for n in _mitres._task_graph.nodes:
        if hasattr(n, "_label"):
            n._label = "FR" + n._label
    return _mitres



from setuptools import setup, find_packages

version = {}
with open("_version.py") as fp:
    exec(fp.read(), version)

setup(
    name="qermit",
    version=version["__version__"],
    python_requires=">=3.6",
    description="error-mitigation framework, an extension to pytket",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    license="CQC Non-Commercial Use Software Licence",
    include_package_data=True,
    packages=find_packages(),
    install_requires=["pytket ~= 0.13.0", "matplotlib", "networkx"],
    classifiers=[
        "Environment :: Console",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: Other/Proprietary License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
    ],
    zip_safe=False,
)
